from typing import Any, Optional

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import wandb
from matplotlib import pyplot as plt

from power_perceiver.consts import BatchKey


class LogNationalPV(pl.Callback):
    """To be used in conjunction with the `NationalPVDataset`. Must be used with a single
    validation data loader worker!

    Attributes:
        _national_pv_accumulator_mw: A pd.DataFrame with columns "actual" and "predicted"
            and index is the target DatetimeIndex (in UTC) of the data.
        _national_batch_idx: The number of complete national "cycles" completed.
    """

    def on_validation_epoch_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        self._national_pv_accumulator_mw = None
        self._national_batch_idx = 0

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[dict[str, torch.Tensor]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        predicted_gsp_power = outputs["predicted_gsp_power_mean"].cpu().detach()  # (example, time)
        actual_gsp_power = batch[BatchKey.gsp].squeeze().cpu()  # shape: (example, time)
        gsp_datetimes = batch[BatchKey.gsp_time_utc].cpu()  # shape: (example, time)
        gsp_capacity_mwp = batch[BatchKey.gsp_capacity_mwp].squeeze().cpu().numpy()  # (example,)
        num_examples = predicted_gsp_power.shape[0]

        gsp_t0_idx = batch[BatchKey.gsp_t0_idx]

        # Select just the forecast timesteps:
        predicted_gsp_power = predicted_gsp_power[:, gsp_t0_idx + 1 :]
        actual_gsp_power = actual_gsp_power[:, gsp_t0_idx + 1 :]
        gsp_datetimes = gsp_datetimes[:, gsp_t0_idx + 1 :]

        for example_idx in range(num_examples):
            dt_for_example = pd.to_datetime(gsp_datetimes[example_idx], unit="s")
            df_for_example = pd.DataFrame(
                {
                    "actual": actual_gsp_power[example_idx],
                    "predicted": predicted_gsp_power[example_idx],
                },
                index=dt_for_example,
            )
            df_for_example *= gsp_capacity_mwp[example_idx]

            if self._national_pv_accumulator_mw is None:
                self._national_pv_accumulator_mw = df_for_example
            elif np.array_equal(dt_for_example, self._national_pv_accumulator_mw.index):
                self._national_pv_accumulator_mw += df_for_example
            else:
                # This is the start of a new set of GSPs. First, log the national PV error:
                self._log_national_pv_error(pl_module)
                if self._national_batch_idx < 12:
                    self._plot_national_pv(pl_module.logger.experiment)
                # And reset:
                self._national_pv_accumulator_mw = df_for_example
                self._national_batch_idx += 1

    def on_validation_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        super().on_validation_epoch_end(trainer, pl_module)
        self._log_national_pv_error(pl_module, on_step=False)

    def _log_national_pv_error(self, logger, on_step: bool = True) -> None:
        self._check_accumulator()
        error_mw = (
            self._national_pv_accumulator_mw["actual"]
            - self._national_pv_accumulator_mw["predicted"]
        )
        logger.log_dict(
            {
                "validation/national_pv_mae_mw": error_mw.abs().mean(),
                "validation/national_pv_mae_mw_2h": error_mw[:4].abs().mean(),
                "validation/national_pv_mbe_mw": error_mw.mean(),
            },
            on_epoch=True,
            on_step=on_step,
        )

    def _plot_national_pv(self, logger) -> None:
        self._check_accumulator()
        fig, ax = plt.subplots()
        ax = self._national_pv_accumulator_mw.plot(ax=ax)
        ax.set_title("National solar PV")
        ax.set_ylabel("MW")
        date = self._national_pv_accumulator_mw.index[0].date()
        ax.set_xlabel(date)
        logger.log({f"validation/national_pv_{self._national_batch_idx}": wandb.Image(fig)})
        plt.close(fig)

    def _check_accumulator(self) -> None:
        assert self._national_pv_accumulator_mw is not None
        assert len(self._national_pv_accumulator_mw) > 0



from typing import Any, Optional

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import pandas as pd
import pytorch_lightning as pl
import torch
import wandb

from power_perceiver.analysis.simple_callback import SimpleCallback
from power_perceiver.consts import BatchKey


def plot_pv_power(
    actual_pv_power: torch.Tensor,
    predicted_pv_power: torch.Tensor,
    actual_gsp_power: torch.Tensor,
    predicted_gsp_power: torch.Tensor,
    example_idx: int,
    pv_datetimes: torch.Tensor,
    gsp_datetimes: torch.Tensor,
    solar_azimuth: torch.Tensor,
    solar_elevation: torch.Tensor,
) -> plt.Figure:
    fig, axes = plt.subplots(nrows=4, sharex=True, sharey=True)
    ax_pv_actual, ax_pv_predicted, ax_gsp, ax_solar = axes

    pv_datetimes = pd.to_datetime(pv_datetimes[example_idx], unit="s")
    gsp_datetimes = pd.to_datetime(gsp_datetimes[example_idx], unit="s")

    # PV
    def _plot_pv(ax, data, title):
        ax.plot(pv_datetimes, data[example_idx].squeeze())
        ax.set_title(title)
        ax.set_ylabel("PV power")

    _plot_pv(ax_pv_actual, actual_pv_power, "Actual PV power")
    _plot_pv(ax_pv_predicted, predicted_pv_power, "Predicted PV power")

    # GSP
    gsp_df = pd.DataFrame(
        {
            "actual_gsp_power": actual_gsp_power[example_idx].squeeze(),
            "predicted_gsp_power": predicted_gsp_power[example_idx].squeeze(),
        },
        index=gsp_datetimes,
    )
    gsp_index_dupes = gsp_df.index.duplicated()
    ax_gsp.plot(
        gsp_df.index[~gsp_index_dupes],
        gsp_df["actual_gsp_power"].loc[~gsp_index_dupes],
        label="Actual",
    )
    ax_gsp.scatter(gsp_df.index, gsp_df["predicted_gsp_power"], alpha=0.8, label="Predicted")
    ax_gsp.set_title("GSP PV power")
    ax_gsp.set_ylabel("GSP PV power")
    ax_gsp.legend()

    # Solar elevation and azimuth
    ax2_solar = ax_solar.twinx()  # Don't share Y axis!
    ax2_solar.plot(pv_datetimes, solar_azimuth[example_idx], label="Solar Azimuth")
    ax2_solar.plot(pv_datetimes, solar_elevation[example_idx], label="Solar Elevation")
    ax2_solar.legend()
    ax2_solar.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M"))
    xlabel = "Time (UTC)"
    if not pd.isnull(pv_datetimes[0]):  # Check this example is not NaT!
        xlabel += "\nDate: " + pv_datetimes[0].date().strftime("%Y-%m-%d")
    ax2_solar.set_xlabel(xlabel)

    return fig


class LogTimeseriesPlots(SimpleCallback):
    def _on_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[dict[str, object]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
        tag: str,
    ) -> None:
        """Called when the training batch ends.

        Args:
            outputs: The output from Model.training_step
            tag: train or validation
        """
        if batch_idx < 4:
            predicted_pv_power = outputs["predicted_pv_power_mean"].cpu().detach()
            actual_pv_power = outputs["actual_pv_power"].cpu().detach()
            # TODO: Generate pv_datetimes upstream and pass it into this function, just like
            # we do with `gsp_time_utc`.
            pv_datetimes = batch[BatchKey.pv_time_utc].cpu()
            n_examples_per_batch = pv_datetimes.shape[0]
            examples_with_gsp_data = [
                example_idx
                for example_idx in range(n_examples_per_batch)
                if batch[BatchKey.gsp_time_utc][example_idx, 0].isfinite()
            ]
            for example_idx in examples_with_gsp_data[:4]:
                fig = plot_pv_power(
                    actual_pv_power=actual_pv_power,
                    predicted_pv_power=predicted_pv_power,
                    actual_gsp_power=outputs["actual_gsp_power"].cpu().detach(),
                    predicted_gsp_power=outputs["predicted_gsp_power_mean"].cpu().detach(),
                    gsp_datetimes=outputs["gsp_time_utc"].cpu().detach(),
                    example_idx=example_idx,
                    pv_datetimes=pv_datetimes,
                    solar_azimuth=batch[BatchKey.hrvsatellite_solar_azimuth].cpu().detach(),
                    solar_elevation=batch[BatchKey.hrvsatellite_solar_elevation].cpu().detach(),
                )
                pl_module.logger.experiment.log(
                    {
                        f"{tag}/pv_power/{batch_idx=};{example_idx=}": wandb.Image(fig),
                        "global_step": trainer.global_step,
                    },
                )
                plt.close(fig)



from numbers import Number
from typing import Union

import numpy as np
import pyproj

# OSGB is also called "OSGB 1936 / British National Grid -- United
# Kingdom Ordnance Survey".  OSGB is used in many UK electricity
# system maps, and is used by the UK Met Office UKV model.  OSGB is a
# Transverse Mercator projection, using 'easting' and 'northing'
# coordinates which are in meters.  See https://epsg.io/27700
OSGB36 = 27700

# WGS84 is short for "World Geodetic System 1984", used in GPS. Uses
# latitude and longitude.
WGS84 = 4326


_osgb_to_lat_lon = pyproj.Transformer.from_crs(crs_from=OSGB36, crs_to=WGS84).transform
_lat_lon_to_osgb = pyproj.Transformer.from_crs(crs_from=WGS84, crs_to=OSGB36).transform


def osgb_to_lat_lon(
    x: Union[Number, np.ndarray], y: Union[Number, np.ndarray]
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Change OSGB coordinates to lat, lon.

    Args:
        x: osgb east-west
        y: osgb north-south
    Return: 2-tuple of latitude (north-south), longitude (east-west).
    """
    return _osgb_to_lat_lon(xx=x, yy=y)


def lat_lon_to_osgb(
    latitude: Union[Number, np.ndarray], longitude: Union[Number, np.ndarray]
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Change lat, lon coordinates to OSGB.

    Return: 2-tuple of OSGB x, y.
    """
    return _lat_lon_to_osgb(xx=latitude, yy=longitude)



from dataclasses import dataclass

import einops
import pytorch_lightning as pl
import torch
from pytorch_msssim import ms_ssim
from torch import nn
from torch.nn import functional as F

from power_perceiver.consts import X_OSGB_MEAN, X_OSGB_STD, Y_OSGB_MEAN, Y_OSGB_STD, BatchKey
from power_perceiver.load_prepared_batches.data_sources.satellite import SAT_MEAN, SAT_STD
from power_perceiver.pytorch_modules.query_generator import GSPQueryGenerator, PVQueryGenerator
from power_perceiver.pytorch_modules.satellite_predictor import XResUNet
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder

NUM_HIST_SAT_IMAGES = 12  # v15 pre-prepared batches use 7
NUM_FUTURE_SAT_IMAGES = 25  # v15 pre-prepared batches use 24
USE_TOPOGRAPHY = True
USE_SUN_POSITION = True
SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS = 24
SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS = 24
SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS = 12
D_MODEL = 256  # Must be divisible by N_HEADS
N_HEADS = 32


def get_osgb_coords_for_coord_conv(batch: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
    """Returns tensor of shape (example, 2, y, x)."""
    y_osgb = batch[BatchKey.hrvsatellite_y_osgb]
    x_osgb = batch[BatchKey.hrvsatellite_x_osgb]

    # Normalise:
    y_osgb = (y_osgb - Y_OSGB_MEAN) / Y_OSGB_STD
    x_osgb = (x_osgb - X_OSGB_MEAN) / X_OSGB_STD

    # Concat:
    return torch.stack((y_osgb, x_osgb), dim=1)


@dataclass(eq=False)
class SatellitePredictor(pl.LightningModule):
    use_coord_conv: bool = False
    crop: bool = False
    use_topography: bool = USE_TOPOGRAPHY
    use_sun_position: bool = USE_SUN_POSITION

    # kwargs to fastai DynamicUnet. See this page for details:
    # https://fastai1.fast.ai/vision.models.unet.html#DynamicUnet
    pretrained: bool = False
    blur_final: bool = True  # Blur final layer. fastai default is True.
    self_attention: bool = True  # Use SA layer at the third block before the end.
    last_cross: bool = True  # Use a cross-connection with the direct input of the model.
    bottle: bool = False  # Bottleneck the last skip connection.

    def __post_init__(self):
        super().__init__()

        self.satellite_predictor = XResUNet(
            img_size=(
                SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS,
                SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS,
            ),
            n_in=(
                NUM_HIST_SAT_IMAGES
                + (2 if self.use_coord_conv else 0)
                + (1 if self.use_topography else 0)
                + (2 if self.use_sun_position else 0)
            ),
            n_out=NUM_FUTURE_SAT_IMAGES,
            pretrained=self.pretrained,
            blur_final=self.blur_final,
            self_attention=self.self_attention,
            last_cross=self.last_cross,
            bottle=self.bottle,
        )

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        data = x[BatchKey.hrvsatellite_actual][
            :, :NUM_HIST_SAT_IMAGES, 0
        ]  # Shape: (example, time, y, x)
        height, width = data.shape[2:]
        assert height == SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS, f"{height=}"
        assert width == SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS, f"{width=}"

        if self.use_coord_conv:
            osgb_coords = get_osgb_coords_for_coord_conv(x)
            data = torch.concat((data, osgb_coords), dim=1)

        if self.use_topography:
            surface_height = x[BatchKey.hrvsatellite_surface_height]
            print(surface_height.shape)
            surface_height = surface_height.unsqueeze(1)  # Add channel dim
            data = torch.concat((data, surface_height), dim=1)

        if self.use_sun_position:
            azimuth_at_t0 = x[BatchKey.hrvsatellite_solar_azimuth][:, NUM_HIST_SAT_IMAGES]
            elevation_at_t0 = x[BatchKey.hrvsatellite_solar_elevation][:, NUM_HIST_SAT_IMAGES]
            sun_pos = torch.stack((azimuth_at_t0, elevation_at_t0), dim=1)  # Shape: (example, 2)
            del azimuth_at_t0, elevation_at_t0
            # Repeat over y and x:
            sun_pos = einops.repeat(sun_pos, "example chan -> example chan y x", y=height, x=width)
            data = torch.concat((data, sun_pos), dim=1)

        assert data.isfinite().all()
        predicted_sat = self.satellite_predictor(data)
        assert predicted_sat.isfinite().all()
        return predicted_sat  # Shape: example, time, y, x

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        predicted_sat = self(batch)
        actual_sat = batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0]

        sat_mse_loss = F.mse_loss(predicted_sat, actual_sat)
        self.log(f"{self.tag}/sat_mse", sat_mse_loss)

        # MS-SSIM. Requires images to be de-normalised:
        actual_sat_denorm = (actual_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        predicted_sat_denorm = (predicted_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        ms_ssim_loss = 1 - ms_ssim(
            predicted_sat_denorm,
            actual_sat_denorm,
            data_range=1023.0,
            size_average=True,  # Return a scalar.
            win_size=3,  # ClimateHack folks used win_size=3.
        )
        self.log(f"{self.tag}/ms_ssim", ms_ssim_loss)
        self.log(f"{self.tag}/ms_ssim+sat_mse", ms_ssim_loss + sat_mse_loss)

        if self.crop_sat_before_sat_predictor_loss:
            # Loss on a central crop:
            # The cropped image has to be larger than 32x32 otherwise ms-ssim complains:
            # "Image size should be larger than 32 due to the 4 downsamplings in ms-ssim"
            sat_mse_loss_crop = F.mse_loss(
                predicted_sat[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                actual_sat[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
            )
            self.log(f"{self.tag}/sat_mse_crop", sat_mse_loss_crop)
            ms_ssim_loss_crop = 1 - ms_ssim(
                predicted_sat_denorm[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                actual_sat_denorm[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                data_range=1023.0,
                size_average=True,  # Return a scalar.
                win_size=3,  # The Illinois ClimateHack folks used win_size=3.
            )
            self.log(f"{self.tag}/ms_ssim_crop", ms_ssim_loss_crop)
            self.log(f"{self.tag}/ms_ssim_crop+sat_mse_crop", ms_ssim_loss_crop + sat_mse_loss_crop)
            sat_loss = ms_ssim_loss_crop + sat_mse_loss_crop
        else:
            sat_loss = ms_ssim_loss + sat_mse_loss

        return dict(
            lost=sat_loss,
        )


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


LEFT_IDX = (SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS // 2) - (
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS // 2
)
RIGHT_IDX = LEFT_IDX + SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS
TOP_IDX = (SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS // 2) - (
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS // 2
)
BOTTOM_IDX = TOP_IDX + SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS


@dataclass(eq=False)
class SatelliteTransformer(nn.Module):
    """Infers a single timestep of PV power and GSP power at a time.

    Currently just uses HRV satellite imagery as the input. In the near future it could also
    use NWP temperature, wind speed & precipitation, and absolute geo position.
    """

    # Params for Perceiver
    # byte_array and query will be automatically padded with zeros to get to d_model.
    # Set d_model to be divisible by `num_heads`.
    d_model: int = D_MODEL
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 8

    def __post_init__(self):
        super().__init__()
        self.hrvsatellite_processor = HRVSatelliteProcessor()

        NUM_PV_SYSTEMS = 1400
        NUM_GSPS = 360
        id_embedding = nn.Embedding(
            num_embeddings=NUM_PV_SYSTEMS + NUM_GSPS,
            embedding_dim=self.pv_system_id_embedding_dim,
        )

        self.pv_query_generator = PVQueryGenerator(
            pv_system_id_embedding=id_embedding,
            num_gsps=NUM_GSPS,
        )

        self.gsp_query_generator = GSPQueryGenerator(gsp_id_embedding=id_embedding)

        self.transformer_encoder = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        # Reshape so each timestep is considered a different example:
        num_examples, num_timesteps = x[BatchKey.pv].shape[:2]
        print(f"{num_examples=} {num_timesteps=}")
        # TODO More timesteps than expected, 43 instead of 31/32
        # Process satellite data and queries:
        pv_query = self.pv_query_generator(x)  # shape: example * time, n_pv_systems, features
        gsp_query = self.gsp_query_generator(
            x,
            include_history=True,
            base_batch_key=BatchKey.gsp_5_min,
            do_reshape_time_as_batch=True,
        )  # shape: example * time, 1, features
        # Some are 32 and some are 31 steps (32*31=996,32*32=1024)
        satellite_data = self.hrvsatellite_processor(x)  # shape: example * time, positions, feats.

        # Pad with zeros if necessary to get up to self.d_model:
        pv_query = maybe_pad_with_zeros(pv_query, requested_dim=self.d_model)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)
        satellite_data = maybe_pad_with_zeros(satellite_data, requested_dim=self.d_model)

        # Prepare inputs for the transformer_encoder:
        print(f"{pv_query.shape=}, {gsp_query.shape=}, {satellite_data.shape=}")
        attn_input = torch.concat((pv_query, gsp_query, satellite_data), dim=1)

        # Mask the NaN GSP and PV queries. True or non-zero value indicates value will be ignored.
        mask = attn_input.isnan().any(dim=2)
        attn_input = attn_input.nan_to_num(0)

        # Pass data into transformer_encoder:
        attn_output = attn_input + self.transformer_encoder(attn_input, src_key_padding_mask=mask)

        # Reshape to (example time element d_model):
        attn_output = einops.rearrange(
            attn_output,
            "(example time) ... -> example time ...",
            example=num_examples,
            time=num_timesteps,
        )

        assert attn_output.isfinite().all()

        # Select the elements of the output which correspond to the query:
        gsp_start_idx = pv_query.shape[1]
        gsp_end_idx = gsp_start_idx + gsp_query.shape[1]
        pv_attn_out = attn_output[:, :, :gsp_start_idx]
        gsp_attn_out = attn_output[:, :, gsp_start_idx:gsp_end_idx]

        return {
            "pv_attn_out": pv_attn_out,  # shape: (example, 5_min_time, n_pv_systems, d_model)
            "gsp_attn_out": gsp_attn_out,  # shape: (example, 5_min_time, 1, d_model)
        }



import datetime
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Optional, Union

import fsspec
import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.consts import BatchKey, Location
from power_perceiver.exceptions import NoPVSystemsInSlice
from power_perceiver.geospatial import lat_lon_to_osgb
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.load_raw.data_sources.raw_data_source import (
    RawDataSource,
    TimeseriesDataSource,
)
from power_perceiver.utils import check_path_exists, datetime64_to_float, select_time_periods

_log = logging.getLogger(__name__)


@dataclass(kw_only=True)
class RawPVDataSource(
    # Surprisingly, Python's class hierarchy is defined right-to-left.
    # So the base class must go on the right.
    TimeseriesDataSource,
    RawDataSource,
):
    """Load PV data directly from the intermediate PV Zarr store.

    Args:
        pv_power_filename:
        pv_metadata_filename:
        roi_height_meters: The height of the region of interest (ROI) when creating examples.
            For PV, we use meters (not pixels) because PV isn't an image.
            Must be at least 1,000 meters.
        roi_width_meters:
        n_pv_systems_per_example: Each example will have exactly this number of PV systems,
            or, if there are zero PV systems in the region (e.g. in northern Scotland),
            then raise `NoPVSystemsInSlice` exception. If there is at least 1 PV system, then
            randomly select PV systems for each example. If there are less PV systems available
            than requested, then randomly sample with duplicates allowed, whilst ensuring all
            available PV systems are used.

    Attributes:
        empty_example: xr.DataArray: An example of the correct shape, but where data and coords
            are all NaNs!
        _data_in_ram: xr.DataArray
            The data is the 5-minutely PV power in Watts.
            Dimension coordinates: time_utc, pv_system_id
            Additional coordinates: x_osgb, y_osgb, capacity_wp
    """

    pv_power_filename: str
    pv_metadata_filename: str
    roi_height_meters: int
    roi_width_meters: int
    n_pv_systems_per_example: int

    # For now, let's assume the PV data is always 5-minutely, even though some PVOutput.org
    # PV systems report data at 15-minutely intervals. For now, let's just interpolate
    # 15-minutely data to 5-minutely. Later (WP3?) we could experiment with giving the model
    # the "raw" (un-interpolated) 15-minutely PV data. See issue #74.
    sample_period_duration: ClassVar[datetime.timedelta] = datetime.timedelta(minutes=5)

    def __post_init__(self):  # noqa: D105
        self._sanity_check_args()
        RawDataSource.__post_init__(self)
        TimeseriesDataSource.__post_init__(self)
        # Load everything into RAM once (at init) rather than in each worker process.
        # This should be faster than loading from disk in every worker!
        self.load_everything_into_ram()
        self.empty_example = self._get_empty_example()

    def _sanity_check_args(self) -> None:
        check_path_exists(self.pv_power_filename)
        check_path_exists(self.pv_metadata_filename)
        assert self.roi_height_meters > 1_000
        assert self.roi_width_meters > 1_000
        assert self.n_pv_systems_per_example > 0

    def load_everything_into_ram(self) -> None:
        """Open AND load PV data into RAM."""
        # Load pd.DataFrame of power and pd.Series of capacities:
        pv_power_watts, pv_capacity_wp, pv_system_row_number = _load_pv_power_watts_and_capacity_wp(
            self.pv_power_filename,
            start_date=self.time_periods.iloc[0]["start_dt"],
            end_date=self.time_periods.iloc[-1]["end_dt"],
        )
        pv_metadata = _load_pv_metadata(self.pv_metadata_filename)
        # Ensure pv_metadata, pv_power_watts, and pv_capacity_wp all have the same set of
        # PV system IDs, in the same order:
        pv_metadata, pv_power_watts = _intersection_of_pv_system_ids(pv_metadata, pv_power_watts)
        pv_capacity_wp = pv_capacity_wp.loc[pv_power_watts.columns]
        pv_system_row_number = pv_system_row_number.loc[pv_power_watts.columns]

        self._data_in_ram = _put_pv_data_into_an_xr_dataarray(
            pv_power_watts=pv_power_watts,
            y_osgb=pv_metadata.y_osgb.astype(np.float32),
            x_osgb=pv_metadata.x_osgb.astype(np.float32),
            capacity_wp=pv_capacity_wp,
            pv_system_row_number=pv_system_row_number,
            t0_idx=self.t0_idx,
            sample_period_duration=self.sample_period_duration,
        )

        self._data_in_ram = select_time_periods(
            xr_data=self.data_in_ram,
            time_periods=self.time_periods,
            dim_name="time_utc",
        )

        # Sanity checks:
        time_utc = pd.DatetimeIndex(self._data_in_ram.time_utc)
        assert time_utc.is_monotonic_increasing
        assert time_utc.is_unique

    def get_osgb_location_for_example(self) -> Location:
        """Get a single random geographical location."""
        raise NotImplementedError(
            "Not planning to implement this just yet. To start with, let's use"
            " `RawSatelliteDataSource` and/or `RawGSPDataSource` to generate locations."
        )

    def _get_spatial_slice(self, xr_data: xr.DataArray, center_osgb: Location) -> xr.DataArray:

        half_roi_width_meters = self.roi_width_meters // 2
        half_roi_height_meters = self.roi_height_meters // 2

        left = center_osgb.x - half_roi_width_meters
        right = center_osgb.x + half_roi_width_meters
        top = center_osgb.y + half_roi_height_meters
        bottom = center_osgb.y - half_roi_height_meters

        # Select data in the region of interest:
        pv_system_id_mask = (
            (left <= xr_data.x_osgb)
            & (xr_data.x_osgb <= right)
            & (xr_data.y_osgb <= top)
            & (bottom <= xr_data.y_osgb)
        )

        selected_data = xr_data.isel(pv_system_id=pv_system_id_mask)

        # Drop any PV systems which have NaN readings at every timestep in the example:
        selected_data = selected_data.dropna(dim="pv_system_id", how="all")

        if len(selected_data.pv_system_id) == 0:
            raise NoPVSystemsInSlice()

        # Interpolate forwards to fill NaNs which follow finite data:
        selected_data = selected_data.interpolate_na(dim="time_utc")
        # Finally, fill any remaining NaNs with zeros. This assumes - probably incorrectly -
        # that any NaNs remaining at this point are at nighttime, and so *should* be zero.
        # TODO: Give the model the "raw" data: Don't `interpolate_na` or `fillna(0)`. See issue #74
        selected_data = selected_data.fillna(0)
        return self._ensure_n_pv_systems_per_example(selected_data)

    def _ensure_n_pv_systems_per_example(self, selected_data: xr.DataArray) -> xr.DataArray:
        """Ensure there are always `self.n_pv_systems_per_example` PV systems."""
        if len(selected_data.pv_system_id) > self.n_pv_systems_per_example:
            # More PV systems are available than we need. Reduce by randomly sampling:
            subset_of_pv_system_ids = self.rng.choice(
                selected_data.pv_system_id,
                size=self.n_pv_systems_per_example,
                replace=False,
            )
            selected_data = selected_data.sel(pv_system_id=subset_of_pv_system_ids)
        elif len(selected_data.pv_system_id) < self.n_pv_systems_per_example:
            # If we just used `choice(replace=True)` then there's a high chance
            # that the output won't include every available PV system but instead
            # will repeat some PV systems at the expense of leaving some on the table.
            # TODO: Don't repeat PV systems. Instead, pad with NaNs and mask the loss. Issue #73.
            n_random_pv_systems = self.n_pv_systems_per_example - len(selected_data.pv_system_id)
            allow_replacement = n_random_pv_systems > len(selected_data.pv_system_id)
            random_pv_system_ids = self.rng.choice(
                selected_data.pv_system_id,
                size=n_random_pv_systems,
                replace=allow_replacement,
            )
            selected_data = xr.concat(
                (selected_data, selected_data.sel(pv_system_id=random_pv_system_ids)),
                dim="pv_system_id",
            )
        return selected_data

    def _post_process(self, xr_data: xr.DataArray) -> xr.DataArray:
        return xr_data / xr_data.capacity_wp

    def _get_empty_example(self) -> xr.DataArray:
        """Return a single example of the correct shape but where data & coords are all NaN."""
        empty_pv_system_ids = np.full(
            shape=self.n_pv_systems_per_example, fill_value=np.NaN, dtype=np.float32
        )
        empty_dt_index = np.full(shape=self.total_seq_length, fill_value=np.NaN)
        empty_dt_index = pd.DatetimeIndex(empty_dt_index)
        empty_pv_power = pd.DataFrame(
            np.NaN, index=empty_dt_index, columns=empty_pv_system_ids
        ).astype(np.float32)
        empty_metadata = pd.Series(np.NaN, index=empty_pv_system_ids).astype(np.float32)
        pv_system_row_number = np.full(
            shape=self.n_pv_systems_per_example, fill_value=np.NaN, dtype=np.float32
        )
        pv_system_row_number = pd.Series(pv_system_row_number, index=empty_pv_system_ids)

        data_array = _put_pv_data_into_an_xr_dataarray(
            pv_power_watts=empty_pv_power,
            y_osgb=empty_metadata,
            x_osgb=empty_metadata,
            capacity_wp=empty_metadata,
            pv_system_row_number=pv_system_row_number,
            t0_idx=self.t0_idx,
            sample_period_duration=self.sample_period_duration,
        )
        return data_array

    @staticmethod
    def to_numpy(xr_data: xr.DataArray) -> NumpyBatch:
        """Return a single example in a `NumpyBatch`.

        We don't bother with the pv_mask in RawPVDataSource. Instead as use the
        rule that PV systems to be ignored will have NaN `pv_id`.
        """
        example: NumpyBatch = {}

        example[BatchKey.pv] = xr_data.values  # PV is normalised in RawPVDataSource._post_process
        example[BatchKey.pv_t0_idx] = xr_data.attrs["t0_idx"]
        example[BatchKey.pv_system_row_number] = xr_data["pv_system_row_number"].values
        example[BatchKey.pv_id] = xr_data["pv_system_id"].values.astype(np.float32)
        example[BatchKey.pv_capacity_wp] = xr_data["capacity_wp"].values
        example[BatchKey.pv_time_utc] = datetime64_to_float(xr_data["time_utc"].values)
        example[BatchKey.pv_x_osgb] = xr_data["x_osgb"].values
        example[BatchKey.pv_y_osgb] = xr_data["y_osgb"].values

        return example


# Adapted from nowcasting_dataset.data_sources.pv.pv_data_source
def _load_pv_power_watts_and_capacity_wp(
    filename: Union[str, Path],
    start_date: Optional[datetime.datetime] = None,
    end_date: Optional[datetime.datetime] = None,
) -> tuple[pd.DataFrame, pd.Series, pd.Series]:
    """Return pv_power_watts, pv_capacity_wp, pv_system_row_number.

    The capacities and pv_system_row_number are computed across the *entire* dataset,
    and so is independent of the `start_date` and `end_date`. This ensures the PV system
    row number and capacities stay constant across training and validation.
    """

    _log.info(f"Loading solar PV power data from {filename} from {start_date=} to {end_date=}.")

    # Load data in a way that will work in the cloud and locally:
    with fsspec.open(filename, mode="rb") as file:
        pv_power_ds = xr.open_dataset(file, engine="h5netcdf")
        pv_capacity_wp = pv_power_ds.max().to_pandas().astype(np.float32)
        pv_power_watts = pv_power_ds.sel(datetime=slice(start_date, end_date)).to_dataframe()
        pv_power_watts = pv_power_watts.astype(np.float32)
        del pv_power_ds

    pv_capacity_wp.index = [np.int32(col) for col in pv_capacity_wp.index]

    # Create pv_system_row_number. We use the index of `pv_capacity_wp` because that includes
    # the PV system IDs for the entire dataset (independent of `start_date` and `end_date`).
    # We use `float32` for the ID because we use NaN to indicate a missing PV system,
    # or that this whole example doesn't include PV.
    all_pv_system_ids = pv_capacity_wp.index
    pv_system_row_number = np.arange(start=0, stop=len(all_pv_system_ids), dtype=np.float32)
    pv_system_row_number = pd.Series(pv_system_row_number, index=all_pv_system_ids)

    _log.info(
        "Before filtering:"
        f" {len(pv_power_watts)} PV power datetimes."
        f" {len(pv_power_watts.columns)} PV power PV system IDs."
    )

    pv_power_watts = pv_power_watts.clip(lower=0, upper=5e7)
    # Convert the pv_system_id column names from strings to ints:
    pv_power_watts.columns = [np.int32(col) for col in pv_power_watts.columns]

    if "passiv" not in filename:
        _log.warning("Converting timezone. ARE YOU SURE THAT'S WHAT YOU WANT TO DO?")
        pv_power_watts = (
            pv_power_watts.tz_localize("Europe/London").tz_convert("UTC").tz_convert(None)
        )

    pv_power_watts = _drop_pv_systems_which_produce_overnight(pv_power_watts)

    # Resample to 5-minutely and interpolate up to 15 minutes ahead.
    # TODO: Issue #74: Give users the option to NOT resample (because Perceiver IO
    # doesn't need all the data to be perfectly aligned).
    pv_power_watts = pv_power_watts.resample("5T").interpolate(method="time", limit=3)
    pv_power_watts.dropna(axis="index", how="all", inplace=True)
    pv_power_watts.dropna(axis="columns", how="all", inplace=True)

    # Drop any PV systems whose PV capacity is too low:
    PV_CAPACITY_THRESHOLD_W = 100
    pv_systems_to_drop = pv_capacity_wp.index[pv_capacity_wp <= PV_CAPACITY_THRESHOLD_W]
    pv_systems_to_drop = pv_systems_to_drop.intersection(pv_power_watts.columns)
    _log.info(
        f"Dropping {len(pv_systems_to_drop)} PV systems because their max power is less than"
        f" {PV_CAPACITY_THRESHOLD_W}"
    )
    pv_power_watts.drop(columns=pv_systems_to_drop, inplace=True)

    # Ensure that capacity and pv_system_row_num use the same PV system IDs as the power DF:
    pv_system_ids = pv_power_watts.columns
    pv_capacity_wp = pv_capacity_wp.loc[pv_system_ids]
    pv_system_row_number = pv_system_row_number.loc[pv_system_ids]

    _log.info(
        "After filtering & resampling to 5 minutes:"
        f" pv_power = {pv_power_watts.values.nbytes / 1e6:,.1f} MBytes."
        f" {len(pv_power_watts)} PV power datetimes."
        f" {len(pv_power_watts.columns)} PV power PV system IDs."
    )

    # Sanity checks:
    assert not pv_power_watts.columns.duplicated().any()
    assert not pv_power_watts.index.duplicated().any()
    assert np.isfinite(pv_capacity_wp).all()
    assert (pv_capacity_wp > 0).all()
    assert np.isfinite(pv_system_row_number).all()
    assert np.array_equal(pv_power_watts.columns, pv_capacity_wp.index)
    return pv_power_watts, pv_capacity_wp, pv_system_row_number


# Adapted from nowcasting_dataset.data_sources.pv.pv_data_source
def _load_pv_metadata(filename: str) -> pd.DataFrame:
    """Return pd.DataFrame of PV metadata.

    Shape of the returned pd.DataFrame for Passiv PV data:
        Index: ss_id (Sheffield Solar ID)
        Columns: llsoacd, orientation, tilt, kwp, operational_at,
            latitude, longitude, system_id, x_osgb, y_osgb
    """
    _log.info(f"Loading PV metadata from {filename}")
    pv_metadata = pd.read_csv(filename, index_col="ss_id").drop(columns="Unnamed: 0")
    _log.info(f"Found {len(pv_metadata)} PV systems in {filename}")

    # drop any systems with no lon or lat:
    pv_metadata.dropna(subset=["longitude", "latitude"], how="any", inplace=True)

    _log.debug(f"Found {len(pv_metadata)} PV systems with locations")

    pv_metadata["x_osgb"], pv_metadata["y_osgb"] = lat_lon_to_osgb(
        latitude=pv_metadata["latitude"], longitude=pv_metadata["longitude"]
    )

    # Remove PV systems outside the geospatial boundary of the satellite data:
    GEO_BOUNDARY_OSGB = {
        "WEST": -238_000,
        "EAST": 856_000,
        "NORTH": 1_222_000,
        "SOUTH": -184_000,
    }

    pv_metadata = pv_metadata[
        (pv_metadata.x_osgb >= GEO_BOUNDARY_OSGB["WEST"])
        & (pv_metadata.x_osgb <= GEO_BOUNDARY_OSGB["EAST"])
        & (pv_metadata.y_osgb <= GEO_BOUNDARY_OSGB["NORTH"])
        & (pv_metadata.y_osgb >= GEO_BOUNDARY_OSGB["SOUTH"])
    ]

    _log.info(f"Found {len(pv_metadata)} PV systems after filtering.")
    return pv_metadata


# From nowcasting_dataset.data_sources.pv.pv_data_source
def _drop_pv_systems_which_produce_overnight(pv_power_watts: pd.DataFrame) -> pd.DataFrame:
    """Drop systems which produce power over night.

    Args:
        pv_power_watts: Un-normalised.
    """
    # TODO: Of these bad systems, 24647, 42656, 42807, 43081, 51247, 59919
    # might have some salvagable data?
    NIGHT_YIELD_THRESHOLD = 0.4
    night_hours = [22, 23, 0, 1, 2]
    pv_power_normalised = pv_power_watts / pv_power_watts.max()
    night_mask = pv_power_normalised.index.hour.isin(night_hours)
    pv_power_at_night_normalised = pv_power_normalised.loc[night_mask]
    pv_above_threshold_at_night = (pv_power_at_night_normalised > NIGHT_YIELD_THRESHOLD).any()
    bad_systems = pv_power_normalised.columns[pv_above_threshold_at_night]
    _log.info(f"{len(bad_systems)} bad PV systems found and removed!")
    return pv_power_watts.drop(columns=bad_systems)


# From nowcasting_dataset.data_sources.pv.pv_data_source
def _intersection_of_pv_system_ids(
    pv_metadata: pd.DataFrame, pv_power: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Only pick PV systems for which we have metadata."""
    pv_system_ids = pv_metadata.index.intersection(pv_power.columns)
    pv_system_ids = np.sort(pv_system_ids)
    pv_power = pv_power[pv_system_ids]
    pv_metadata = pv_metadata.loc[pv_system_ids]
    return pv_metadata, pv_power


def _put_pv_data_into_an_xr_dataarray(
    pv_power_watts: pd.DataFrame,
    y_osgb: pd.Series,
    x_osgb: pd.Series,
    capacity_wp: pd.Series,
    pv_system_row_number: pd.Series,
    t0_idx: int,
    sample_period_duration: datetime.timedelta,
) -> xr.DataArray:
    """Convert to an xarray DataArray.

    Args:
        pv_power_watts: pd.DataFrame where the columns are PV systems (and the column names are
            ints), and the index is UTC datetime.
        x_osgb: The x location. Index = PV system ID ints.
        y_osgb: The y location. Index = PV system ID ints.
        capacity_wp: The max power output of each PV system in Watts. Index = PV system ID ints.
        pv_system_row_number: The integer position of the PV system in the metadata.
            Used to create the PV system ID embedding.
    """
    # Sanity check!
    pv_system_ids = pv_power_watts.columns
    for series in (x_osgb, y_osgb, capacity_wp, pv_system_row_number):
        assert np.array_equal(series.index, pv_system_ids, equal_nan=True)

    data_array = xr.DataArray(
        data=pv_power_watts.values,
        coords=(("time_utc", pv_power_watts.index), ("pv_system_id", pv_power_watts.columns)),
        name="pv_power_watts",
    ).astype(np.float32)

    data_array = data_array.assign_coords(
        x_osgb=("pv_system_id", x_osgb),
        y_osgb=("pv_system_id", y_osgb),
        capacity_wp=("pv_system_id", capacity_wp),
        pv_system_row_number=("pv_system_id", pv_system_row_number),
    )
    data_array.attrs["t0_idx"] = t0_idx
    # Sample period duration is required so PVDownsample transform knows by how much
    # to change the pv_t0_idx:
    data_array.attrs["sample_period_duration"] = sample_period_duration
    return data_array



"""Utilities to help with unit testing."""

import logging
import os
from pathlib import Path

from gcsfs import GCSFileSystem

from power_perceiver.consts import REMOTE_PATH_FOR_DATA_FOR_UNIT_TESTS

INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING = (0, 1)
_log = logging.getLogger(__name__)


def get_filename_of_batch_of_data_and_maybe_download(data_source_name: str, batch_idx: int) -> Path:
    """If the data is already downloaded locally then return the filename.

    The testing data is too large to include in the git repo so we publish
    a small amount of data to a public Google Cloud Storage bucket.

    If the data is not already downloaded then download and return the filename.

    Args:
        data_source_name (str): e.g. "hrvsatellite"
        batch_idx (int): 0 or 1.

    Returns:
        Path: The full path to the local data.
    """
    assert (
        batch_idx in INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING
    ), f"batch_idx must be {INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING}, not {batch_idx}"
    local_data_path = get_path_of_local_data_for_testing()
    local_path_for_data_source = local_data_path / data_source_name
    if not local_path_for_data_source.exists():
        _log.info(f"Creating path for local testing data: {local_path_for_data_source}")
        os.makedirs(local_path_for_data_source)
    nc_filename = f"{batch_idx:06d}.nc"
    local_batch_filename = local_path_for_data_source / nc_filename
    if not local_batch_filename.exists():
        remote_batch_filename = REMOTE_PATH_FOR_DATA_FOR_UNIT_TESTS / data_source_name / nc_filename
        _log.info(f"Downloading {remote_batch_filename}")
        fs = GCSFileSystem()
        fs.get(str(remote_batch_filename), str(local_batch_filename))
    return local_batch_filename


def download_batches_for_data_source_if_necessary(data_source_name: str) -> list[Path]:
    """Download batches locally, if necessary.

    Args:
        data_source_name (str):

    Returns:
        list[Path]: A list of the Paths pointing to the local files.
    """
    filenames = []
    for batch_idx in INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING:
        filename = get_filename_of_batch_of_data_and_maybe_download(
            data_source_name=data_source_name,
            batch_idx=batch_idx,
        )
        filenames.append(filename)
    return filenames


def _get_path_of_power_perceiver_package() -> Path:
    import power_perceiver

    return Path(power_perceiver.__file__).parents[1]


def get_path_of_local_data_for_testing() -> Path:
    return _get_path_of_power_perceiver_package() / "data_for_testing"


def get_path_of_local_topo_data() -> Path:
    return _get_path_of_power_perceiver_package() / "data" / "europe_dem_2km_osgb.tif"



from typing import Any, Optional

import pytorch_lightning as pl


class SimpleCallback(pl.Callback):
    """A simple abstract class which redirects `on_train_batch_end` and `on_validation_batch_end`
    to a single method, `_on_batch_end`."""

    def _on_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[dict[str, object]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
        tag: str,
    ) -> None:
        """Called when the batch ends.

        Args:
            outputs: The output from Model.training_step
            tag: 'train' or 'validation'
        """
        raise NotImplementedError  # MUST BE IMPLEMENTED BY SUBCLASSES!

    def on_train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[dict[str, object]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self._on_batch_end(
            trainer=trainer,
            pl_module=pl_module,
            outputs=outputs,
            batch=batch,
            batch_idx=batch_idx,
            dataloader_idx=dataloader_idx,
            tag="train",
        )

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[dict[str, object]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self._on_batch_end(
            trainer=trainer,
            pl_module=pl_module,
            outputs=outputs,
            batch=batch,
            batch_idx=batch_idx,
            dataloader_idx=dataloader_idx,
            tag="validation",
        )



import numpy as np
import xarray as xr

from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import (
    BatchKey,
    NumpyBatch,
    PreparedDataSource,
)

ELEVATION_MEAN = 37.4
ELEVATION_STD = 12.7
AZIMUTH_MEAN = 177.7
AZIMUTH_STD = 41.7


class Sun(PreparedDataSource):
    def process_before_transforms(self, dataset: xr.Dataset) -> xr.Dataset:
        # None of this will be necessary once this is implemented:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/635

        # Drop redundant coordinates (these are redundant because they
        # just repeat the contents of each *dimension*):
        dataset = dataset.drop_vars(["example", "time_index"])

        # Rename coords to be more explicit about exactly what some coordinates hold:
        dataset = dataset.rename_vars({"time": "time_utc"})

        # Rename dimensions. Standardize on the singular (time, channel, etc.).
        # Remove redundant "index" from the dim name. These are *dimensions* so,
        # by definition, they are indicies!
        dataset = dataset.rename_dims({"time_index": "time"})

        dataset = dataset.set_coords("time_utc")
        return dataset

    @staticmethod
    def to_numpy(dataset: xr.Dataset) -> NumpyBatch:
        """This is called from Dataset.__getitem__.

        Sets `BatchKey.solar_azimuth` and `BatchKey.solar_elevation`.

        This processes this modality's xr.Dataset, to convert the xr.Dataset
        into a dictionary mapping BatchKeys to numpy arrays, as documented
        in the BatchKey class.
        """
        batch: NumpyBatch = {}

        solar_azimuth = dataset["azimuth"].astype(np.float32).values
        solar_azimuth -= AZIMUTH_MEAN
        solar_azimuth /= AZIMUTH_STD
        batch[BatchKey.hrvsatellite_solar_azimuth] = solar_azimuth

        solar_elevation = dataset["elevation"].astype(np.float32).values
        solar_elevation -= ELEVATION_MEAN
        solar_elevation /= ELEVATION_STD
        batch[BatchKey.hrvsatellite_solar_elevation] = solar_elevation

        return batch



import numpy as np
import pytest

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources import PV, HRVSatellite
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.np_batch_processor import EncodeSpaceTime
from power_perceiver.np_batch_processor.encode_space_time import compute_fourier_features
from power_perceiver.testing import get_path_of_local_data_for_testing


def test_fourier_features():
    test_array = np.arange(start=-2, stop=2, step=0.01, dtype=np.float32).reshape((1, -1))
    fourier_features = compute_fourier_features(test_array)
    assert fourier_features.shape == (1, 400, 8)
    assert fourier_features.dtype == np.float32


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_encode_space_time():
    dataset = PreparedDataset(
        data_path=get_path_of_local_data_for_testing(),
        data_loaders=[HRVSatellite(), PV()],
        np_batch_processors=[EncodeSpaceTime()],
    )
    np_batch = dataset[0]

    for batch_key in (
        BatchKey.hrvsatellite_time_utc_fourier,
        BatchKey.hrvsatellite_x_osgb_fourier,
        BatchKey.hrvsatellite_y_osgb_fourier,
        BatchKey.pv_time_utc_fourier,
        BatchKey.pv_x_osgb_fourier,
        BatchKey.pv_y_osgb_fourier,
    ):
        assert batch_key in np_batch, f"{batch_key.name} is missing from np_batch!"

    assert np_batch[BatchKey.hrvsatellite_x_osgb_fourier].shape == (32, 64, 64, 8)



import logging
from typing import Any, Optional

import matplotlib
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import wandb

from power_perceiver.analysis.simple_callback import SimpleCallback
from power_perceiver.consts import BatchKey
from power_perceiver.pytorch_modules.mixture_density_network import plot_probs

_log = logging.getLogger(__name__)


def plot_pv_power(
    actual_pv_power: torch.Tensor,  # shape: example, time, n_pv_systems
    predicted_pv_power: torch.Tensor,
    actual_gsp_power: torch.Tensor,
    predicted_gsp_power: torch.Tensor,
    example_idx: int,
    pv_datetimes: torch.Tensor,
    gsp_datetimes: torch.Tensor,
    actual_satellite: torch.Tensor,  # shape: example, time, channels, y, x
    surface_height: torch.Tensor,
    predicted_pv_power_mean: torch.Tensor,
    predicted_gsp_power_mean: torch.Tensor,
    gsp_id: torch.Tensor,
    pv_id: torch.Tensor,
    nwp: torch.Tensor,
    nwp_target_time_utc: torch.Tensor,
    nwp_channel_names: torch.Tensor,
) -> plt.Figure:
    fig, axes = plt.subplots(nrows=3, ncols=4)
    axes = np.array(axes).flatten()

    pv_datetimes = pd.to_datetime(pv_datetimes[example_idx], unit="s")
    gsp_datetimes = pd.to_datetime(gsp_datetimes[example_idx], unit="s")
    nwp_datetimes = pd.to_datetime(nwp_target_time_utc[example_idx], unit="s")

    if pd.isnull(pv_datetimes[0]):
        # This example has no PV data.
        return fig

    legend_kwargs = dict(framealpha=0.4, fontsize="x-small")

    # PV
    for pv_idx, ax in enumerate(axes[:-4]):
        plot_probs(
            ax=ax,
            network_output=predicted_pv_power[example_idx, :, pv_idx],
            left=pv_datetimes[0],
            right=pv_datetimes[-1],
        )
        ax.set_title("PV power for {:.0f}".format(pv_id[example_idx, pv_idx]))
        try:
            ax.plot(pv_datetimes, actual_pv_power[example_idx, :, pv_idx], label="Actual PV")
        except Exception as e:
            raise e.__class__(
                f"{pv_datetimes.shape=}; {gsp_datetimes.shape=};"
                f" {actual_pv_power.shape=}; {predicted_pv_power.shape=}"
            ) from e
        ax.plot(
            pv_datetimes, predicted_pv_power_mean[example_idx, :, pv_idx], label="Mean prediction"
        )
        ax.legend(**legend_kwargs)

    # GSP
    ax_gsp = axes[-4]
    plot_probs(
        ax=ax_gsp,
        network_output=predicted_gsp_power[example_idx].squeeze(),
        left=gsp_datetimes[0],
        right=gsp_datetimes[-1],
    )
    ax_gsp.plot(gsp_datetimes, actual_gsp_power[example_idx], label="Actual")
    ax_gsp.plot(gsp_datetimes, predicted_gsp_power_mean[example_idx], label="Mean prediction")
    ax_gsp.set_title("GSP PV power for {:.0f}".format(gsp_id[example_idx]))
    ax_gsp.set_xlabel(pv_datetimes[0].date().strftime("%Y-%m-%d"))
    ax_gsp.legend(**legend_kwargs)

    # NWP
    ax_nwp = axes[-3]
    ax_nwp.set_title("NWP (mean over x and y)")
    ax_nwp_twin = ax_nwp.twiny()
    ax_nwp_twin.plot(
        nwp_datetimes,
        nwp[example_idx].mean(dim=(-1, -2)),
        label=nwp_channel_names,
    )
    ax_nwp_twin.legend(loc="upper right", **legend_kwargs)
    ax_nwp.tick_params(axis="x", which="both", bottom=False, labelbottom=False)
    ax_nwp_twin.tick_params(
        axis="x", which="both", bottom=True, labelbottom=True, top=False, labeltop=False
    )

    # Format all the timeseries plots (PV and GSP and NWP)
    for ax in np.concatenate((axes[:-2], [ax_nwp_twin])):
        ax.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M"))
        ax.tick_params(axis="x", labelsize="x-small")

    # Satellite
    sat_axes = [ax.twinx().twiny() for ax in axes[-2:]]

    actual_satellite = actual_satellite[example_idx, :, 0]
    sat_axes[0].imshow(actual_satellite[0])
    sat_axes[0].set_xlabel("First actual satellite image")

    sat_axes[1].imshow(actual_satellite[-1])
    sat_axes[1].set_xlabel("Last actual satellite image")

    def _turn_off_ticks(ax: matplotlib.axes.Axes):
        ax.tick_params(
            axis="both",
            which="both",
            bottom=False,
            top=False,
            left=False,
            right=False,
            labelbottom=False,
            labelleft=False,
            labeltop=False,
            labelright=False,
        )

    for ax in axes[-2:]:
        _turn_off_ticks(ax)

    for ax in sat_axes:
        _turn_off_ticks(ax)

    return fig


class LogProbabilityTimeseriesPlots(SimpleCallback):
    def _on_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[dict[str, object]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
        tag: str,
    ) -> None:
        """Called when the training batch ends.

        Args:
            outputs: The output from Model.training_step
            tag: train or validation
        """
        if batch_idx in range(0, 128, 16):
            predicted_pv_power = outputs["predicted_pv_power"].cpu().detach()
            actual_pv_power = outputs["actual_pv_power"].cpu().detach()
            pv_datetimes = outputs["pv_time_utc"].cpu().detach()
            n_examples_per_batch = pv_datetimes.shape[0]
            examples_with_gsp_data = [
                example_idx
                for example_idx in range(n_examples_per_batch)
                if batch[BatchKey.gsp_time_utc][example_idx].isfinite().all()
            ]
            example_idx = examples_with_gsp_data[0]
            fig = plot_pv_power(
                actual_pv_power=actual_pv_power,
                predicted_pv_power=predicted_pv_power,
                predicted_pv_power_mean=outputs["predicted_pv_power_mean"].cpu().detach(),
                actual_gsp_power=outputs["actual_gsp_power"].cpu().detach(),
                predicted_gsp_power=outputs["predicted_gsp_power"].cpu().detach(),
                predicted_gsp_power_mean=outputs["predicted_gsp_power_mean"].cpu().detach(),
                gsp_datetimes=outputs["gsp_time_utc"].cpu().detach(),
                example_idx=example_idx,
                pv_datetimes=pv_datetimes,
                actual_satellite=batch[BatchKey.hrvsatellite_actual].cpu(),
                surface_height=batch[BatchKey.hrvsatellite_surface_height].cpu(),
                pv_id=batch[BatchKey.pv_id].cpu(),
                gsp_id=batch[BatchKey.gsp_id].squeeze().cpu(),
                nwp=batch[BatchKey.nwp].cpu(),
                nwp_target_time_utc=batch[BatchKey.nwp_target_time_utc].cpu(),
                nwp_channel_names=batch[BatchKey.nwp_channel_names],
            )
            pl_module.logger.experiment.log(
                {f"{tag}/pv_power_probs/{batch_idx=};{example_idx=}": wandb.Image(fig)}
            )
            plt.close(fig)



from dataclasses import dataclass

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch


@dataclass
class SaveT0Time:
    """Save the T0 time_utc_fourier for PV and GSP data.

    This is useful to ensure the model always knows the t0 timestep,
    even after the timesteps have been subsampled.
    """

    def __call__(self, np_batch: NumpyBatch) -> NumpyBatch:
        pv_t0_idx = np_batch[BatchKey.pv_t0_idx]
        gsp_t0_idx = np_batch[BatchKey.gsp_t0_idx]
        hrvsatellite_t0_idx = np_batch[BatchKey.hrvsatellite_t0_idx]

        np_batch[BatchKey.pv_time_utc_fourier_t0] = np_batch[BatchKey.pv_time_utc_fourier][
            :, pv_t0_idx
        ]
        np_batch[BatchKey.gsp_time_utc_fourier_t0] = np_batch[BatchKey.gsp_time_utc_fourier][
            :, gsp_t0_idx
        ]
        np_batch[BatchKey.hrvsatellite_time_utc_fourier_t0] = np_batch[
            BatchKey.hrvsatellite_time_utc_fourier
        ][:, hrvsatellite_t0_idx]
        return np_batch



from dataclasses import dataclass

import einops
import torch
from torch import nn

from power_perceiver.consts import PV_SPACER_LEN, SATELLITE_SPACER_LEN, BatchKey
from power_perceiver.pytorch_modules.utils import get_spacer_tensor


@dataclass(eq=False)
class NWPProcessor(nn.Module):
    n_channels: int  # Number of NWP channels.
    channel_id_dim: int = 16

    def __post_init__(self):
        super().__init__()
        self.learnable_channel_ids = nn.Parameter(torch.randn(self.n_channels, self.channel_id_dim))

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        """Return a byte array ready for Perceiver.

        Args:
            x: A batch with at least these BatchKeys:
                nwp
                nwp_target_time_utc_fourier
                nwp_init_time_utc_fourier

        Returns:
            tensor of shape (example, time * channel, feature).
        """

        # Patch all positions into a single element:
        nwp = x[BatchKey.nwp]  # Shape: (example, time, channel, y, x)
        nwp = einops.rearrange(nwp, "example time channel y x -> example time channel (y x)")
        assert nwp.shape[2] == self.n_channels

        # Repeat time fourier over all channels
        def _get_time(batch_key: BatchKey) -> torch.Tensor:
            return einops.repeat(
                x[batch_key],
                "example time features -> example time channel features",
                channel=self.n_channels,
            )

        time_fourier = _get_time(
            BatchKey.nwp_target_time_utc_fourier
        )  # (example, time, n_features)
        init_time_fourier = _get_time(BatchKey.nwp_init_time_utc_fourier)

        # Repeat learnt embedding ID for each channel:
        channel_ids = einops.repeat(
            self.learnable_channel_ids,
            "channel features -> example time channel features",
            example=nwp.shape[0],
            time=nwp.shape[1],
        )

        # Sun position:
        def _get_solar_position(batch_key: BatchKey) -> torch.Tensor:
            return einops.repeat(
                x[batch_key],
                "example time -> example time channel 1",
                channel=self.n_channels,
            )

        solar_azimuth = _get_solar_position(BatchKey.nwp_target_time_solar_azimuth)
        solar_elevation = _get_solar_position(BatchKey.nwp_target_time_solar_elevation)

        y_fourier_dummy = x_fourier_dummy = torch.zeros_like(time_fourier)
        # length = 20 for satellite +
        satellite_and_pv_spacer = get_spacer_tensor(
            template=time_fourier,
            length=SATELLITE_SPACER_LEN + PV_SPACER_LEN,
        )

        # Concatenate on the feature dimension (the last dim):
        nwp_query = torch.concat(
            (
                time_fourier,
                init_time_fourier,
                solar_azimuth,
                solar_elevation,
                y_fourier_dummy,
                x_fourier_dummy,
                satellite_and_pv_spacer,
                channel_ids,
                nwp,
            ),
            dim=-1,
        )

        return einops.rearrange(
            nwp_query, "example time channel features -> example (time channel) features"
        )



import logging
from functools import partial

import torch
from fastai.layers import Mish
from fastai.vision.learner import create_unet_model
from fastai.vision.models.xresnet import xse_resnext50
from torch import nn

_log = logging.getLogger(__name__)


class XResUNet(nn.Module):
    """Predict future satellite images using a U-Net.

    This model was developed by the Illinois team in ClimateHack.AI in early 2022.
    This model won the ClimateHack.AI competition :). See
    https://github.com/jmather625/climatehack
    """

    def __init__(self, arch=None, **kwargs):
        """Create a DynamicUnet.

        See this page for a description of the kwargs:
        https://fastai1.fast.ai/vision.models.unet.html#DynamicUnet
        """
        _log.info(f"kwargs for XResUNet = {kwargs}")
        super().__init__()
        if arch is None:
            arch = partial(xse_resnext50, act_cls=Mish, sa=True)
        self.model = create_unet_model(arch=arch, **kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)



from typing import Any, Optional

import matplotlib.pyplot as plt
import pandas as pd
import pytorch_lightning as pl
import torch
import wandb

from power_perceiver.analysis.simple_callback import SimpleCallback
from power_perceiver.consts import BatchKey


def plot_satellite(
    actual_sat: torch.Tensor,  # shape: (example, time, y, x)
    predicted_sat: torch.Tensor,
    example_idx: torch.Tensor,
    sat_datetimes: torch.Tensor,
    num_timesteps: int = 8,
    interval: int = 3,
) -> plt.Figure:

    datetimes = pd.to_datetime(sat_datetimes[example_idx], unit="s")
    actual_sat = actual_sat[example_idx]
    predicted_sat = predicted_sat[example_idx]

    # Use the same "exposure" for all plots:
    vmin = min(actual_sat.min(), predicted_sat.min())
    vmax = max(actual_sat.max(), predicted_sat.max())

    fig, axes = plt.subplots(nrows=2, ncols=num_timesteps)
    for ax, tensor, title in zip(axes, (actual_sat, predicted_sat), ("actual", "predicted")):
        for i in range(num_timesteps):
            timestep_idx = int(i * interval)
            dt = datetimes[timestep_idx]
            dt = dt.strftime("%Y-%m-%d %H:%M")
            image = tensor[timestep_idx]
            ax[i].imshow(image, vmin=vmin, vmax=vmax)
            ax[i].set_title(f"{title} {timestep_idx=}\n{dt}")

    return fig


class LogSatellitePlots(SimpleCallback):
    def _on_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[dict[str, object]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
        tag: str,
    ) -> None:
        """Called when the training batch ends.

        Args:
            outputs: The output from Model.training_step
            tag: train or validation
        """
        if batch_idx < 4:
            predicted_sat = outputs["predicted_sat"].cpu().detach()
            actual_sat = outputs["actual_sat"].cpu().detach()
            sat_datetimes = batch[BatchKey.hrvsatellite_time_utc].cpu()
            for example_idx in range(4):
                fig = plot_satellite(
                    actual_sat=actual_sat,
                    predicted_sat=predicted_sat,
                    example_idx=example_idx,
                    sat_datetimes=sat_datetimes,
                )
                pl_module.logger.experiment.log(
                    {
                        f"{tag}/satellite/{batch_idx=};{example_idx=}": wandb.Image(fig),
                    },
                )
                plt.close(fig)

                # Plot surface height:
                fig, ax = plt.subplots()
                ax.imshow(batch[BatchKey.hrvsatellite_surface_height][example_idx].cpu())
                pl_module.logger.experiment.log(
                    {f"{tag}/surface_height/{batch_idx=};{example_idx=}": wandb.Image(fig)}
                )
                plt.close(fig)



from dataclasses import dataclass
from typing import Any, Optional

import numpy as np
import pytorch_lightning as pl
import sklearn.manifold
import torch
import wandb
from matplotlib import pyplot as plt
from torch import nn

from power_perceiver.analysis.simple_callback import SimpleCallback
from power_perceiver.consts import BatchKey


def plot_tsne_of_pv_system_id_embedding(
    batch: dict[BatchKey, torch.Tensor],
    pv_system_id_embedding: nn.Embedding,
    device,
) -> plt.Figure:
    """Plot the t-SNE of the embedding for all PV systems in the batch."""
    pv_system_row_numbers_for_all_examples = []
    num_examples = batch[BatchKey.pv_system_row_number].shape[0]
    for example_idx in range(num_examples):
        if BatchKey.pv_mask in batch:
            # Data is from Prepared PV data loader.
            pv_mask = batch[BatchKey.pv_mask][example_idx]
        else:
            # Data is from RawPVDataLoader.
            pv_mask = batch[BatchKey.pv_id][example_idx].isfinite()
        row_numbers_for_example = batch[BatchKey.pv_system_row_number][example_idx]
        row_numbers_for_example = row_numbers_for_example[pv_mask]
        if len(row_numbers_for_example) > 0:
            row_numbers_for_example = row_numbers_for_example.tolist()
            pv_system_row_numbers_for_all_examples.extend(row_numbers_for_example)

    pv_system_row_numbers_for_all_examples = np.unique(pv_system_row_numbers_for_all_examples)
    NUM_GSPS = 360
    pv_system_row_numbers_for_all_examples += NUM_GSPS
    pv_system_row_numbers_for_all_examples = torch.from_numpy(
        pv_system_row_numbers_for_all_examples
    ).to(device=device)

    all_gsp_ids = torch.arange(start=1, end=NUM_GSPS, device=device)

    all_ids = torch.concat((all_gsp_ids, pv_system_row_numbers_for_all_examples))
    all_ids = torch.nan_to_num(all_ids).int()

    pv_id_embedding = pv_system_id_embedding(all_ids)
    pv_id_embedding = pv_id_embedding.detach().cpu()

    tsne = sklearn.manifold.TSNE(n_components=2, init="pca", learning_rate="auto")
    tsne = tsne.fit_transform(pv_id_embedding)

    fig, ax = plt.subplots()
    # Use different colours for GSP and PV
    colors = (["#384fd1"] * (NUM_GSPS - 1)) + (
        ["#eb7005"] * len(pv_system_row_numbers_for_all_examples)
    )
    ax.scatter(x=tsne[:, 0], y=tsne[:, 1], alpha=0.6, c=colors)
    ax.set_title(
        "t-SNE of PV system ID embedding for all examples in batch."
        " GSPs are blue. PV systems are orange"
    )

    return fig


@dataclass
class LogTSNEPlot(SimpleCallback):
    query_generator_name: str = "decoder_query_generator"

    def __post_init__(self):
        super().__init__()

    def _on_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[dict[str, object]],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
        tag: str,
    ) -> None:
        """Called when the training batch ends.

        Args:
            outputs: The output from Model.training_step
            tag: train or validation
        """
        if batch_idx == 0:
            query_generator = multilayer_getattr(pl_module, self.query_generator_name)
            fig = plot_tsne_of_pv_system_id_embedding(
                batch=batch,
                pv_system_id_embedding=query_generator.pv_system_id_embedding,
                device=pl_module.device,
            )
            # Need to convert to image to avoid bug in matplotlib to plotly conversion
            pl_module.logger.experiment.log({f"{tag}/tsne": wandb.Image(fig)})
            plt.close(fig)


def multilayer_getattr(obj, name: str):
    """Like getattr, except `name` can have dots in it."""
    attr_names = name.split(".")
    attr = obj
    for attr_name in attr_names:
        attr = getattr(attr, attr_name)
    return attr



import logging

import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import (
    BatchKey,
    NumpyBatch,
    PreparedDataSource,
)
from power_perceiver.utils import datetime64_to_float

_log = logging.getLogger(__name__)


class PV(PreparedDataSource):
    sample_period_duration = pd.Timedelta("5 min")

    def process_before_transforms(self, dataset: xr.Dataset) -> xr.Dataset:
        # None of this will be necessary once this is implemented:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/630

        # Drop redundant coordinates (these are redundant because they
        # just repeat the contents of each *dimension*):
        dataset = dataset.drop_vars(["example", "id_index", "time_index"])

        # Rename coords to be more explicit about exactly what some coordinates hold:
        # Note that, in v15 of the dataset, the keys are incorrectly named
        # power_mw and capacity_mwp, even though the power and capacity are both in watts.
        # See https://github.com/openclimatefix/nowcasting_dataset/issues/530
        dataset = dataset.rename_vars(
            {
                "time": "time_utc",
                "power_mw": "power_w",
                "capacity_mwp": "capacity_wp",
                "id": "pv_system_id",
            }
        )

        # Rename dimensions. Standardize on the singular (time, channel, etc.).
        # Remove redundant "index" from the dim name. These are *dimensions* so,
        # by definition, they are indicies!
        dataset = dataset.rename_dims(
            {
                "time_index": "time",
                "id_index": "pv_system",
            }
        )

        # Setting coords won't be necessary once this is fixed:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/627
        dataset = dataset.set_coords(
            ["time_utc", "pv_system_id", "pv_system_row_number", "y_osgb", "x_osgb"]
        )

        dataset = dataset.transpose("example", "time", "pv_system")

        # PV power
        # Note that some capacities are 0. This will be fixed upstream in:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/622
        dataset["power_normalised"] = dataset["power_w"] / dataset["capacity_wp"]
        dataset["power_normalised"] = dataset["power_normalised"].astype(np.float32)

        # Compute mask of valid PV data.
        # The mask will be a bool DataArray of shape (batch_size, n_pv_systems).
        # In v15, some capacities are 0. This will be fixed upstream in:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/622
        valid_pv_capacity = dataset["capacity_wp"] > 0
        # A NaN ID value is the "official" way to indicate a missing or deselected PV system.
        valid_pv_id = np.isfinite(dataset["pv_system_id"])
        valid_pv_power = np.isfinite(dataset["power_normalised"]).all(dim="time")
        pv_mask = valid_pv_capacity & valid_pv_id & valid_pv_power
        assert pv_mask.any(), "No valid PV systems!"
        dataset["pv_mask"] = pv_mask

        # Apply the mask. In particular, this is especially important so we set the
        # osgb coordinates to NaN for missing PV systems. In v15, the osgb coordinates
        # are zero for missing PV systems, which confuses `EncodeSpaceTime`!
        dataset = apply_pv_mask(dataset)

        # PV spatial coords are float64 in v15. This will be fixed in:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/624
        for dataset_key in ("x_osgb", "y_osgb"):
            dataset[dataset_key] = dataset[dataset_key].astype(np.float32)
        dataset.attrs["t0_idx"] = self.t0_idx
        return dataset

    @staticmethod
    def to_numpy(dataset: xr.Dataset) -> NumpyBatch:
        """This is called from Dataset.__getitem__.

        This processes this modality's xr.Dataset, to convert the xr.Dataset
        into a dictionary mapping BatchKeys to numpy arrays, as documented
        in the BatchKey class.
        """
        batch: NumpyBatch = {}

        # PV power
        # Note that some capacities are 0. This will be fixed upstream in:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/622
        batch[BatchKey.pv] = dataset["power_normalised"].values
        batch[BatchKey.pv_t0_idx] = dataset.attrs["t0_idx"]

        # In v15 of the dataset, `pv_system_row_number` is int64. This will be fixed in:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/624
        batch[BatchKey.pv_system_row_number] = dataset["pv_system_row_number"].values.astype(
            np.int32
        )

        # id is float64 in v15 of the dataset. This will be fixed upstream in:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/624
        batch[BatchKey.pv_id] = dataset["pv_system_id"].values.astype(np.float32)
        batch[BatchKey.pv_capacity_wp] = dataset["capacity_wp"].values

        batch[BatchKey.pv_mask] = dataset["pv_mask"].values

        # Coordinates
        batch[BatchKey.pv_time_utc] = datetime64_to_float(dataset["time_utc"].values)
        for batch_key, dataset_key in (
            (BatchKey.pv_x_osgb, "x_osgb"),
            (BatchKey.pv_y_osgb, "y_osgb"),
        ):
            batch[batch_key] = dataset[dataset_key].values

        return batch


def apply_pv_mask(dataset: xr.Dataset) -> xr.Dataset:
    """Apply mask to whole dataset (to set everything masked to NaN)"""
    # Coords for missing PV systems are set to 0 in v15. This is dangerous! So set to NaN.
    # We need to set them to NaN so `np.nanmax()` does the right thing in `EncodeSpaceTime`
    # This won't be necessary after this issue is closed:
    # https://github.com/openclimatefix/nowcasting_dataset/issues/625

    dataset = dataset.where(dataset.pv_mask)
    # `Dataset.where` sets `pv_mask` to zero. We want `pv_mask` to be a bool array:
    dataset["pv_mask"] = dataset.pv_mask.fillna(0).astype(bool)
    # `Dataset.where` *only* operates on data variables, so we need to manually mask the coords:
    for coord_name in ["pv_system_id", "x_osgb", "y_osgb"]:
        dataset[coord_name] = dataset[coord_name].where(dataset.pv_mask)
    return dataset



from copy import deepcopy

import pytest

from power_perceiver.consts import BatchKey
from power_perceiver.load_raw.raw_dataset import RawDataset

N_EXAMPLES_PER_BATCH = 16
SAT_HEIGHT_IN_PIXELS = 128
SAT_WIDTH_IN_PIXELS = 256
SAT_N_EXPECTED_TIMESTEPS = 37  # 12 steps of history + 1 for t0 + 24 of forecast


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_init(raw_dataset_with_sat_only: RawDataset):
    assert len(raw_dataset_with_sat_only._all_t0_periods_per_combo) == 0
    assert len(raw_dataset_with_sat_only._t0_datetimes_per_combo_for_epoch) == 0


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_per_worker_init(raw_dataset_with_sat_only: RawDataset):
    dataset = deepcopy(raw_dataset_with_sat_only)
    dataset.per_worker_init(worker_id=1)


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
@pytest.mark.parametrize(
    "raw_dataset_str", ["raw_dataset_with_sat_only", "raw_dataset_with_sat_only_and_gsp_pv_sat"]
)
def test_iter(raw_dataset_str: str, request):
    # `getfixturevalue` trick from https://stackoverflow.com/a/64348247/732596
    raw_dataset: RawDataset = request.getfixturevalue(raw_dataset_str)
    dataset = deepcopy(raw_dataset)
    dataset.per_worker_init(worker_id=1)
    for np_batch in dataset:
        break

    for key, expected_shape in (
        (
            BatchKey.hrvsatellite_actual,
            (
                N_EXAMPLES_PER_BATCH,
                SAT_N_EXPECTED_TIMESTEPS,
                1,
                SAT_HEIGHT_IN_PIXELS,
                SAT_WIDTH_IN_PIXELS,
            ),
        ),
        (
            BatchKey.hrvsatellite_time_utc,
            (
                N_EXAMPLES_PER_BATCH,
                SAT_N_EXPECTED_TIMESTEPS,
            ),
        ),
        (
            BatchKey.hrvsatellite_y_osgb,
            (N_EXAMPLES_PER_BATCH, SAT_HEIGHT_IN_PIXELS, SAT_WIDTH_IN_PIXELS),
        ),
        (
            BatchKey.hrvsatellite_x_osgb,
            (N_EXAMPLES_PER_BATCH, SAT_HEIGHT_IN_PIXELS, SAT_WIDTH_IN_PIXELS),
        ),
        (
            BatchKey.hrvsatellite_y_geostationary,
            (
                N_EXAMPLES_PER_BATCH,
                SAT_HEIGHT_IN_PIXELS,
            ),
        ),
        (
            BatchKey.hrvsatellite_x_geostationary,
            (
                N_EXAMPLES_PER_BATCH,
                SAT_WIDTH_IN_PIXELS,
            ),
        ),
    ):
        value = np_batch[key]
        assert (
            value.shape == expected_shape
        ), f"{key.name=} has shape {value.shape}, not {expected_shape}"


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_dataset_with_sat_only_and_gsp_pv_sat(raw_dataset_with_sat_only_and_gsp_pv_sat: RawDataset):
    pass



from dataclasses import dataclass


@dataclass
class DataSource:
    pre_processors: list
    post_processors: list


class ProductionPVDataSource(DataSource):
    pass


class SatelliteDataSource(DataSource):
    pass


class RemoveBadPVSystems:
    pass


class FillNighttimePVWithNaNs:
    pass


@dataclass
class NowcastingProductionDataset:
    data_sources: list


class FourierEncodeRelativeSpaceTime:
    pass


class SunPosition:
    pass


class PVNet:
    pass


@dataclass
class RunModel:
    model: PVNet


class DisableForecastsRunAtNight:
    pass


@dataclass
class ClipPredictedPowerToZeroAtNight:
    elevation_threshold_degrees: int


class SaveForecastsToDB:
    pass


def run_data_pipeline(pipeline: list, space_time_location_selector):
    pass


class Normalize:
    pass


class SaveBatchesToDisk:
    pass


class NormalizePV:
    pass


class EncodePVSystemsWithGSPID:
    pass


class Select15MinSatellite:
    pass


class NormalizeSatellite:
    pass


class PatchSatellite:
    pass


class EncodePixelsWithGSPID:
    pass


def now():
    pass


"""
Design objectives:

- Make the code as modular as possible.
- Make it super-easy to see all the data processing steps in one glance.
- Make it easy to configure different data processing for different models.
- Share code across training and production:
  - Data processors
  - DataSources that load from Zarr.

Definitely haven't fully thought this through yet!
"""

sat_xr_dataset = OpenSatelliteZarr(sat_zarr_path)
# OpenSatelliteZarr returns the entire, lazily-opened satellite dataset,
# and does any renaming and transposing necessary to get the data into the
# "standard" form. I've put some thought into what that "standard" form could be. Please see:
# https://github.com/openclimatefix/power_perceiver/blob/main/power_perceiver/load_raw/data_sources/raw_satellite_data_source.py#L147

pv_xr_dataset = LoadPVFromDB(db_credentials)
# Returns an xr.DataArray of all the PV data in RAM.
# Replace `LoadPVFromDB` with LoadPVFromNetCDF when training. Both output data in exactly the same shape.
pv_xr_dataset = remove_bad_pv_systems(pv_xr_dataset)

############# GET LIST OF ALL AVAILABLE T0 DATETIMES, ACROSS MODALITIES ##################
# This only needs to be done once (not for every example).
# `sat_t0_datetimes` and `pv_t0_datetimes` could just be `pd.DatetimeIndex` objects.
sat_t0_datetimes = get_all_available_t0_datetimes(
    sat_xr_dataset,
    history_duration=config.satellite.history_duration,
    forecast_duration=config.satellite.forecast_duration,
)
pv_t0_datetimes = get_all_available_t0_datetimes(
    pv_xr_dataset,
    history_duration=config.pv.history_duration,
    forecast_duration=config.pv.forecast_duration,
)

# Find all available t0 datetimes across all modalities:
available_t0_datetimes = sat_t0_datetimes.intersection(pv_t0_datetimes)

# The PV and Satellite pipes both need to be told which locations to load during training:
# This sends the following instructions to the data loaders:
# Batch 1: Load super-batch 0 into RAM. When done, yield example at t0 & location x,y.
#          Then, in the background, load super-batch 1.
# Batch 2: Yield example at t0, location x, y
# <last batch of the epoch>: Switch to super-batch 1. (Delete super-batch 0 from RAM).
# In the background, load super-batch 2. Yield example at t0, location x, y.
# Replace T0PickerForSuperBatchLoader when in production, if when training and you don't want to load super batches.
t0 = T0PickerForSuperBatchLoader(
    available_t0_datetimes=available_t0_datetimes,
    t0_freq="15T",
    locations_per_timestep=4,
)
t0_for_pv, t0_for_sat = Forker(t0, num_instances=2, buffer_size=0)

# Select random locations for each example. Center each example on a random PV system:
location = LocationPicker(locations=pv_xr_dataset)

# -------------- PV DataPipe ------------
pv_pipe = SelectTimeSlice(
    pv_xr_dataset, t0_for_pv
)  # Forwards the t0_datetime & location to subsequent steps.
# Maybe it starts to populate an Example object, where Example.pv is the PV xr.DataArray; and Example.pv_t0_time_utc, Example.center_lat, and Example.center_lon specify the location?
# New: Split the history and forecast into two separate objects. e.g. BatchML.inputs.pv and BatchML.targets.pv?
# Or maybe that adds more complexity???
pv_pipe = SelectPVSystemsWithinRegion(
    pv_pipe,
    location=location,
    roi_width_km=config.pv.roi.width_km,
    roi_height_km=config.pv.roi.height_km,
)
pv_pipe = FillNighttimePVWithNaNs(pv_pipe)
pv_pipe = InterpolateMissingPV(pv_pipe)
pv_pipe = NormalizePV(pv_pipe)
pv_pipe = SunPosition(pv_pipe)

# Validation:
# In this example, don't check for NaNs in PV, because the model expects missing PV to be represented as NaNs.
pv_pipe = CheckStatisticalProperties(pv_pipe)
pv_pipe = CheckPVShape(pv_pipe)

# -------------- Satellite DataPipe ---------
sat_pipe = LoadSuperBatchIntoRAM(sat_xr_dataset, t0_for_sat)  # Optional. Useful during training.
# LoadSuperBatchIntoRAM speeds up training by caching a subset of the on-disk dataset into RAM,
# specified by SpateTimeLocationPicker.
sat_pipe = SelectTimeSlice(sat_pipe)
sat_pipe = SelectSpatialSlice(sat_pipe)
sat_pipe = NormalizeSatellite(sat_pipe)
sat_pipe = PatchSatellite(sat_pipe)
sat_pipe = SunPosition(sat_pipe)

# Validation:
sat_pipe = CheckForNaNs(sat_pipe)
sat_pipe = CheckStatisticalProperties(sat_pipe)
sat_pipe = CheckSatShape(sat_pipe)

# -------------- Merge & process -----------------
main_pipe = MergeBatchML(pv_pipe, sat_pipe)

main_pipe = FourierEncodeRelativeSpaceTime(main_pipe)

# Saving batches to disk is entirely optional during production,
# where it may be useful for debugging:
# During training, SaveBatchesToDisk can be used to save pre-prepared batches.
# And then `LoadBatchesFromDisk()` can be used to load batches into this pipeline.
main_pipe = SaveBatchesToDisk(main_pipe)

# Run ML model:
main_pipe = RunModel(main_pipe, model=PVNet())

# Post-process output from model:
main_pipe = ClipPredictedPowerToZeroAtNight(main_pipe, elevation_threshold_degrees=5)
main_pipe = SaveForecastsToDB(main_pipe)



from typing import Callable, Iterable

import numpy as np
import pandas as pd
import pytest

from power_perceiver.consts import PV_SYSTEM_AXIS, PV_TIME_AXIS, BatchKey
from power_perceiver.load_prepared_batches.data_sources import (
    GSP,
    NWP,
    PV,
    HRVSatellite,
    PreparedDataSource,
)
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.np_batch_processor import EncodeSpaceTime
from power_perceiver.testing import (
    INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING,
    download_batches_for_data_source_if_necessary,
    get_path_of_local_data_for_testing,
)
from power_perceiver.transforms.pv import PVPowerRollingWindow
from power_perceiver.transforms.satellite import PatchSatellite
from power_perceiver.xr_batch_processor import SelectPVSystemsNearCenterOfImage

_DATA_SOURCES_TO_DOWNLOAD = (HRVSatellite.name, PV.name, GSP.name, NWP.name)
BATCH_SIZE = 32
N_PV_TIMESTEPS = 31
N_PV_SYSTEMS_PER_EXAMPLE = 128


def setup_module():
    for data_source_name in _DATA_SOURCES_TO_DOWNLOAD:
        download_batches_for_data_source_if_necessary(data_source_name)


@pytest.mark.parametrize(
    argnames=["max_n_batches_per_epoch", "expected_n_batches"],
    argvalues=[(None, len(INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING)), (1, 1)],
)
def test_init(max_n_batches_per_epoch: int, expected_n_batches: int):
    dataset = PreparedDataset(
        data_path=get_path_of_local_data_for_testing(),
        data_loaders=[PV(history_duration=pd.Timedelta("90 min"))],
        max_n_batches_per_epoch=max_n_batches_per_epoch,
    )
    assert dataset.n_batches == expected_n_batches
    assert len(dataset) == expected_n_batches


@pytest.mark.parametrize(
    argnames=["data_loader", "expected_batch_keys"],
    argvalues=[
        (HRVSatellite(history_duration=pd.Timedelta("30 min")), [BatchKey.hrvsatellite_actual]),
        (PV(history_duration=pd.Timedelta("90 min")), [BatchKey.pv, BatchKey.pv_system_row_number]),
    ],
)
def test_dataset_with_single_data_source(
    data_loader: PreparedDataSource, expected_batch_keys: Iterable[BatchKey]
):
    dataset = PreparedDataset(
        data_path=get_path_of_local_data_for_testing(),
        data_loaders=[data_loader],
    )
    np_batch = dataset[0]
    assert len(np_batch) > 0
    assert isinstance(np_batch, dict)
    assert all([isinstance(key, BatchKey) for key in np_batch])
    assert all(
        [
            isinstance(value, np.ndarray)
            for key, value in np_batch.items()
            if "t0_idx" not in key.name
        ]
    )
    for batch_key in expected_batch_keys:
        assert batch_key in np_batch, f"{batch_key} not in np_data!"


def _check_pv_batch(
    np_batch: NumpyBatch,
    expected_batch_size: int = BATCH_SIZE,
    expected_n_pv_timesteps: int = N_PV_TIMESTEPS,
    expected_n_pv_systems_per_example: int = N_PV_SYSTEMS_PER_EXAMPLE,
) -> None:

    pv_batch = np_batch[BatchKey.pv]
    assert len(pv_batch.shape) == 3
    assert pv_batch.shape[0] == expected_batch_size
    assert pv_batch.shape[PV_TIME_AXIS] == expected_n_pv_timesteps
    assert pv_batch.shape[PV_SYSTEM_AXIS] == expected_n_pv_systems_per_example

    pv_system_row_number = np_batch[BatchKey.pv_system_row_number]
    assert len(pv_system_row_number.shape) == 2
    assert pv_system_row_number.shape[0] == expected_batch_size
    assert pv_system_row_number.shape[1] == expected_n_pv_systems_per_example

    pv_mask = np_batch[BatchKey.pv_mask]
    assert len(pv_mask.shape) == 2
    assert pv_mask.shape[0] == expected_batch_size
    assert pv_mask.shape[1] == expected_n_pv_systems_per_example

    # Select valid PV systems, and check the timeseries are not NaN.
    assert pv_mask.any(), "No valid PV systems!"
    pv_is_finite = np.isfinite(pv_batch).all(axis=PV_TIME_AXIS)
    assert ((~pv_is_finite | ~pv_mask) == ~pv_mask).all()


def test_select_pv_systems_near_center_of_image():
    xr_batch_processors = [SelectPVSystemsNearCenterOfImage()]

    dataset = PreparedDataset(
        data_path=get_path_of_local_data_for_testing(),
        data_loaders=[
            HRVSatellite(history_duration=pd.Timedelta("30 min")),
            PV(history_duration=pd.Timedelta("90 min")),
        ],
        xr_batch_processors=xr_batch_processors,
    )
    np_batch = dataset[0]
    for key, value in np_batch.items():
        if isinstance(value, np.ndarray):
            print(f"{key}: {value.shape}")
        else:
            print(f"{key}: {value}")

    # Batch 0 has 1 example with no PV systems within the region of interest.
    _check_pv_batch(np_batch, expected_batch_size=BATCH_SIZE - 1)

    # Batch 1 has 4 examples with no PV systems within the region of interest.
    np_batch = dataset[1]
    _check_pv_batch(np_batch, expected_batch_size=BATCH_SIZE - 4)


@pytest.mark.parametrize(argnames="transforms", argvalues=[None, [PVPowerRollingWindow()]])
def test_pv(transforms: Iterable[Callable]):
    pv_data_loader = PV(transforms=transforms, history_duration=pd.Timedelta("90 min"))
    dataset = PreparedDataset(
        data_path=get_path_of_local_data_for_testing(),
        data_loaders=[pv_data_loader],
    )
    assert len(dataset) == len(INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING)
    for batch_idx in range(len(INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING)):
        np_batch = dataset[batch_idx]
        for key, value in np_batch.items():
            if isinstance(value, np.ndarray):
                print(f"{key}: {value.shape}")
            else:
                print(f"{key}: {value}")
        _check_pv_batch(np_batch)


def test_all_data_loaders_and_all_transforms():
    dataset = PreparedDataset(
        data_path=get_path_of_local_data_for_testing(),
        data_loaders=[
            HRVSatellite(transforms=[PatchSatellite()], history_duration=pd.Timedelta("30 min")),
            PV(transforms=[PVPowerRollingWindow()], history_duration=pd.Timedelta("90 min")),
        ],
        # xr_batch_processors=[SelectPVSystemsNearCenterOfImage()],
        np_batch_processors=[EncodeSpaceTime()],
    )
    for batch_idx in range(len(INDEXES_OF_PUBLICLY_AVAILABLE_BATCHES_FOR_TESTING)):
        np_batch = dataset[batch_idx]
        if batch_idx == 1:
            # Batch 1 has 4 examples with no PV systems within the region of interest.
            expected_batch_size = BATCH_SIZE - 4
        else:
            expected_batch_size = BATCH_SIZE - 1
        _check_pv_batch(np_batch, expected_batch_size=expected_batch_size)
        # shape is (example, time, channel, y, x, patch)
        assert np_batch[BatchKey.hrvsatellite_actual].shape == (
            expected_batch_size,
            31,
            1,
            16,
            16,
            16,
        )
        assert np_batch[BatchKey.hrvsatellite_x_osgb].shape == (expected_batch_size, 16, 16)
        assert np_batch[BatchKey.hrvsatellite_y_osgb].shape == (expected_batch_size, 16, 16)
        assert np_batch[BatchKey.hrvsatellite_x_osgb_fourier].shape == (
            expected_batch_size,
            16,
            16,
            8,
        )
        assert np_batch[BatchKey.hrvsatellite_y_osgb_fourier].shape == (
            expected_batch_size,
            16,
            16,
            8,
        )



import datetime
import logging
from dataclasses import dataclass
from typing import ClassVar, Optional, Sequence

import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.load_raw.data_sources.raw_data_source import (
    RawDataSource,
    SpatialDataSource,
    TimeseriesDataSource,
    ZarrDataSource,
)
from power_perceiver.time import date_summary_str
from power_perceiver.utils import datetime64_to_float, is_sorted, select_time_periods

_log = logging.getLogger(__name__)


# Means and std computed with
# nowcasting_dataset/scripts/compute_stats_from_batches.py
# using v15 training batches on 2021-11-24.
NWP_MEAN = {
    "t": 285.7799539185846,
    "dswrf": 294.6696933986283,
    "prate": 3.6078121378638696e-05,
    "r": 75.57106712435926,
    "sde": 0.0024915961594965614,
    "si10": 4.931356852411006,
    "vis": 22321.762918384553,
    "lcc": 47.90454236572895,
    "mcc": 44.22781694449808,
    "hcc": 32.87577371914454,
}

NWP_STD = {
    "t": 5.017000766747606,
    "dswrf": 233.1834250473355,
    "prate": 0.00021690701537950742,
    "r": 15.705370079694358,
    "sde": 0.07560040052148084,
    "si10": 2.664583614352396,
    "vis": 12963.802514945439,
    "lcc": 40.06675870700349,
    "mcc": 41.927221148316384,
    "hcc": 39.05157559763763,
}

NWP_CHANNEL_NAMES = tuple(NWP_STD.keys())


def _to_data_array(d):
    return xr.DataArray(
        [d[key] for key in NWP_CHANNEL_NAMES], coords={"channel": list(NWP_CHANNEL_NAMES)}
    ).astype(np.float32)


NWP_MEAN = _to_data_array(NWP_MEAN)
NWP_STD = _to_data_array(NWP_STD)


@dataclass(kw_only=True)
class RawNWPDataSource(
    # Surprisingly, Python's class hierarchy is defined right-to-left.
    # So the base class must go on the right.
    ZarrDataSource,
    TimeseriesDataSource,
    SpatialDataSource,
    RawDataSource,
):
    """Load NWPs direction from the intermediate NWP Zarr store.

    In particular, loads the "UKV" NWP from the UK Met Office, downloaded from CEDA.
    "UKV" stands for "United Kingdom Variable", and it the UK Met Office's high-res deterministic
    NWP for the UK. All the NWP variables are represented in the `variable` dimension within
    the UKV DataArray.

    x is left-to-right.
    y is top-to-bottom.

    Init args:
        zarr_path:
        roi_height_pixels:
        roi_width_pixels:
        history_duration: datetime.timedelta
        forecast_duration: datetime.timedelta
        time_periods: pd.DataFrame
        y_coarsen:
        x_coarsen: Downsample by taking the mean across this number of pixels.
        channels: The NWP forecast parameters to load. If None then don't filter.
            See:  http://cedadocs.ceda.ac.uk/1334/1/uk_model_data_sheet_lores1.pdf
            All of these params are "instant" (i.e. a snapshot at the target time,
            not accumulated over some time period).  The available params are:
                cdcb  : Height of lowest cloud base > 3 oktas, in meters above surface.
                lcc   : Low-level cloud cover in %.
                mcc   : Medium-level cloud cover in %.
                hcc   : High-level cloud cover in %.
                sde   : Snow depth in meters.
                hcct  : Height of convective cloud top, meters above surface.
                        WARNING: hcct has NaNs where there are no clouds forecast to exist!
                dswrf : Downward short-wave radiation flux in W/m^2 (irradiance) at surface.
                dlwrf : Downward long-wave radiation flux in W/m^2 (irradiance) at surface.
                h     : Geometrical height, meters.
                t     : Air temperature at 1 meter above surface in Kelvin.
                r     : Relative humidty in %.
                dpt   : Dew point temperature in Kelvin.
                vis   : Visibility in meters.
                si10  : Wind speed in meters per second, 10 meters above surface.
                wdir10: Wind direction in degrees, 10 meters above surface.
                prmsl : Pressure reduce to mean sea level in Pascals.
                prate : Precipitation rate at the surface in kg/m^2/s.
    """

    y_coarsen: int
    x_coarsen: int
    channels: Optional[Sequence[str]] = None

    # Attributes which are intended to be set for the whole class:
    sample_period_duration: ClassVar[datetime.timedelta] = datetime.timedelta(hours=1)
    duration_between_nwp_inits: ClassVar[datetime.timedelta] = datetime.timedelta(hours=3)
    needs_to_load_subset_into_ram: ClassVar[bool] = True
    _y_dim_name: ClassVar[str] = "y_osgb"
    _x_dim_name: ClassVar[str] = "x_osgb"
    _time_dim_name: ClassVar[str] = "init_time_utc"

    def __post_init__(self):  # noqa: D105
        if self.channels:
            chans_not_in_channel_names = set(self.channels) - set(NWP_CHANNEL_NAMES)
            assert len(chans_not_in_channel_names) == 0, (
                f"{len(chans_not_in_channel_names)} requested channel name(s) not in"
                f" NWP_CHANNEL_NAMES! {chans_not_in_channel_names=}; {self.channels=};"
                f" {NWP_CHANNEL_NAMES=}"
            )
        RawDataSource.__post_init__(self)
        SpatialDataSource.__post_init__(self)
        TimeseriesDataSource.__post_init__(self)
        ZarrDataSource.__post_init__(self)
        self.empty_example = self._get_empty_example()

    def _get_empty_example(self) -> xr.DataArray:
        target_time_utc = pd.DatetimeIndex([np.NaN] * self.total_seq_length)
        channels = self.channels if self.channels else NWP_CHANNEL_NAMES
        n_channels = len(channels)
        data = np.full(
            shape=(
                self.total_seq_length,
                n_channels,
                self.roi_height_pixels,
                self.roi_width_pixels,
            ),
            fill_value=np.NaN,
            dtype=np.float32,
        )
        y = np.full(
            shape=self.roi_height_pixels,
            fill_value=np.NaN,
            dtype=np.float32,
        )
        x = np.full(
            shape=self.roi_width_pixels,
            fill_value=np.NaN,
            dtype=np.float32,
        )
        data_array = xr.DataArray(
            data,
            coords=(
                ("target_time_utc", target_time_utc),
                ("channel", list(channels)),
                ("y_osgb", y),
                ("x_osgb", x),
            ),
        )
        data_array.attrs["t0_idx"] = self.t0_idx
        return data_array

    def per_worker_init(self, *args, **kwargs) -> None:  # noqa: D102
        super().per_worker_init(*args, **kwargs)
        self.open()

    def open(self) -> None:
        """
        Lazily open NWP data (this only loads metadata. It doesn't load all the data!)

        We don't want to call `open_sat_data` in __init__.
        If we did that, then we couldn't copy RawNWPDataSource
        instances into separate processes.  Instead,
        call open() _after_ creating separate processes.
        """
        self._data_on_disk = open_nwp(zarr_path=self.zarr_path)
        if self.channels is not None:
            self._data_on_disk = self._data_on_disk.sel(channel=list(self.channels))
        self.channels = self.data_on_disk.channel.values

        # Check the x and y coords are sorted.
        assert is_sorted(self.data_on_disk[self._y_dim_name][::-1])
        assert is_sorted(self.data_on_disk[self._x_dim_name])

        # Sub-select data:
        _log.info("Before any selection: " + date_summary_str(self.data_on_disk.init_time_utc))

        # Select only the timesteps we want:
        self._data_on_disk = select_time_periods(
            xr_data=self.data_on_disk,
            time_periods=self.time_periods,
            dim_name=self._time_dim_name,
        )
        # Downsample spatially:
        self._data_on_disk = self.data_on_disk.coarsen(
            y_osgb=self.y_coarsen,
            x_osgb=self.x_coarsen,
            boundary="trim",
        ).mean()
        _log.info("After filtering: " + date_summary_str(self.data_on_disk.init_time_utc))

    @property
    def datetime_index(self) -> pd.DatetimeIndex:
        """Return a complete list of all available datetimes.

        We need to return the `target_time_utc` (the times the NWPs are _about_).
        The `target_time` is the `init_time_utc` plus the forecast horizon `step`.
        `step` is an array of timedeltas, so we can just add `init_time_utc` to `step`.
        """
        # Only include steps that are at most `total_duration` apart.
        data = self.data_on_disk.sel(step=slice(None, self.total_duration))
        target_times = data.init_time_utc + data.step
        target_times = target_times.values.flatten()
        target_times = np.unique(target_times)
        target_times = np.sort(target_times)
        target_times = pd.DatetimeIndex(target_times)
        return target_times

    def _get_time_slice(
        self, xr_data: xr.DataArray, t0_datetime_utc: datetime.datetime
    ) -> xr.DataArray:
        """Select a timeslice from `xr_data`.

        The NWP for each example covers a contiguous timespan running from `start_dt` to `end_dt`.
        The first part of the timeseries [`start_dt`, `t0`] is the 'recent history'.  The second
        part of the timeseries (`t0`, `end_dt`] is the 'future'. For each timestep in the
        recent history [`start`, `t0`], get predictions produced by the freshest NWP initialisation
        to each target_time_utc. For the future (`t0`, `end`], use the NWP initialised most
        recently to t0.

        The returned data does not include an `example` dimension.
        """
        start_dt_ceil = self._get_start_dt_ceil(t0_datetime_utc)
        end_dt_ceil = self._get_end_dt_ceil(t0_datetime_utc)

        target_times = pd.date_range(start_dt_ceil, end_dt_ceil, freq=self.sample_period_duration)

        # Get the most recent NWP initialisation time for each target_time_hourly.
        init_times = xr_data.sel(init_time_utc=target_times, method="pad").init_time_utc.values

        # Find the NWP init time for just the 'future' portion of the example.
        init_time_t0 = init_times[self.t0_idx]

        # For the 'future' portion of the example, replace all the NWP
        # init times with the NWP init time most recent to t0.
        init_times[self.t0_idx :] = init_time_t0

        steps = target_times - init_times

        # We want one timestep for each target_time_hourly (obviously!) If we simply do
        # nwp.sel(init_time=init_times, step=steps) then we'll get the *product* of
        # init_times and steps, which is not what # we want! Instead, we use xarray's
        # vectorized-indexing mode by using a DataArray indexer.  See the last example here:
        # https://docs.xarray.dev/en/latest/user-guide/indexing.html#more-advanced-indexing
        coords = {"target_time_utc": target_times}
        init_time_indexer = xr.DataArray(init_times, coords=coords)
        step_indexer = xr.DataArray(steps, coords=coords)
        time_slice = xr_data.sel(step=step_indexer, init_time_utc=init_time_indexer)

        self._sanity_check_time_slice(time_slice, "target_time_utc", t0_datetime_utc)
        return time_slice

    def _post_process(self, xr_data: xr.DataArray) -> xr.DataArray:
        xr_data = xr_data - NWP_MEAN
        xr_data = xr_data / NWP_STD
        return xr_data

    def _convert_t0_time_periods_to_periods_to_load(
        self, subset_of_contiguous_t0_time_periods: pd.DataFrame
    ) -> pd.DataFrame:
        """Convert t0_time_periods back into the complete time periods we want to load.

        Need to stretch the `time_periods` because the nwp init_time is only at
        `duration_between_nwp_inits` intervals.
        """
        time_periods = super()._convert_t0_time_periods_to_periods_to_load(
            subset_of_contiguous_t0_time_periods
        )
        # Get the init time that's most recent to each start_dt.
        time_periods["start_dt"] = self.data_on_disk.sel(
            init_time_utc=time_periods.start_dt.values, method="pad"
        ).init_time_utc
        return time_periods

    @staticmethod
    def to_numpy(xr_data: xr.DataArray) -> NumpyBatch:
        """Convert xarray to numpy batch.

        But note that this is actually just returns *one* example (not a whole batch!)
        """
        example: NumpyBatch = {}

        example[BatchKey.nwp] = xr_data.values
        example[BatchKey.nwp_t0_idx] = xr_data.attrs["t0_idx"]
        target_time = xr_data.target_time_utc.values
        example[BatchKey.nwp_target_time_utc] = datetime64_to_float(target_time)
        example[BatchKey.nwp_channel_names] = xr_data.channel.values
        example[BatchKey.nwp_step] = (xr_data.step.values / np.timedelta64(1, "h")).astype(np.int64)
        example[BatchKey.nwp_init_time_utc] = datetime64_to_float(xr_data.init_time_utc.values)

        for batch_key, dataset_key in (
            (BatchKey.nwp_y_osgb, "y_osgb"),
            (BatchKey.nwp_x_osgb, "x_osgb"),
        ):
            example[batch_key] = xr_data[dataset_key].values

        return example


def open_nwp(zarr_path: str) -> xr.DataArray:
    """
    Open The NWP data.

    Args:
        zarr_path: zarr_path must start with 'gs://' if it's on GCP.

    Returns: NWP data.
    """
    _log.debug("Opening NWP data: %s", zarr_path)
    nwp = xr.open_dataset(
        zarr_path,
        engine="zarr",
        consolidated=True,
        mode="r",
        chunks="auto",
    )

    ukv = nwp["UKV"]
    del nwp

    ukv = ukv.transpose("init_time", "step", "variable", "y", "x")
    ukv = ukv.rename(
        {"init_time": "init_time_utc", "variable": "channel", "y": "y_osgb", "x": "x_osgb"}
    )

    # y_osgb and x_osgb are int64 on disk.
    for coord_name in ("y_osgb", "x_osgb"):
        ukv[coord_name] = ukv[coord_name].astype(np.float32)

    # Sanity checks.
    assert ukv.y_osgb[0] > ukv.y_osgb[1], "UKV must run from top-to-bottom."
    time = pd.DatetimeIndex(ukv.init_time_utc)
    assert time.is_unique
    assert time.is_monotonic_increasing

    return ukv



from dataclasses import InitVar, dataclass
from typing import Optional, Sequence

import einops
import torch
from torch import nn

from power_perceiver.consts import SATELLITE_SPACER_LEN, BatchKey
from power_perceiver.pytorch_modules.utils import get_spacer_tensor
from power_perceiver.utils import assert_num_dims

# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
# for why we set `eq=False`


@dataclass(eq=False)
class PVQueryGenerator(nn.Module):
    """Create a query from the locations of the PV systems."""

    # This must be an InitVar because PyTorch does not allow modules to be
    # assigned before super().__init__()
    pv_system_id_embedding: InitVar[nn.Embedding]
    num_gsps: int = 360  # Used to make sure PV IDs don't clash with GSP IDs!

    def __post_init__(self, pv_system_id_embedding):
        super().__init__()
        self.pv_system_id_embedding = pv_system_id_embedding

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        """The returned tensor is of shape (example * time, n_pv_systems, query_dim)."""
        timeless_x = reshape_time_as_batch(
            x,
            batch_keys=(
                BatchKey.pv,
                BatchKey.pv_solar_azimuth,
                BatchKey.pv_solar_elevation,
                BatchKey.pv_time_utc_fourier,
            ),
            set_to_zero_after_t0_idx=x[BatchKey.pv_t0_idx],
            batch_keys_to_zero_out=[BatchKey.pv],
        )

        n_timesteps, n_pv_systems = x[BatchKey.pv].shape[1:]

        y_fourier = x[BatchKey.pv_y_osgb_fourier]  # (example, n_pv_systems, fourier_features)
        x_fourier = x[BatchKey.pv_x_osgb_fourier]

        pv_system_row_number = x[BatchKey.pv_system_row_number]  # (example, n_pv_systems)
        pv_system_row_number = pv_system_row_number + self.num_gsps
        pv_system_embedding = self.pv_system_id_embedding(pv_system_row_number.nan_to_num(0).int())

        time_fourier = timeless_x[
            BatchKey.pv_time_utc_fourier
        ]  # shape: ((example * time) features)
        assert_num_dims(time_fourier, 2)
        time_fourier_t0 = x[BatchKey.pv_time_utc_fourier_t0]  # shape: (example, features)

        # Repeat y_fourier, x_fourier, and pv_system_embedding across each timestep:
        y_fourier = y_fourier.repeat_interleave(repeats=n_timesteps, dim=0)
        x_fourier = x_fourier.repeat_interleave(repeats=n_timesteps, dim=0)
        pv_system_embedding = pv_system_embedding.repeat_interleave(repeats=n_timesteps, dim=0)
        time_fourier_t0 = time_fourier_t0.repeat_interleave(repeats=n_timesteps, dim=0)

        # Repeat across all PV systems:
        time_fourier = einops.repeat(
            time_fourier,
            "example features -> example n_pv_systems features",
            n_pv_systems=n_pv_systems,
        )
        time_fourier_t0 = einops.repeat(
            time_fourier_t0,
            "example features -> example n_pv_systems features",
            n_pv_systems=n_pv_systems,
        )

        # Reshape solar features to: (example, n_pv_systems, 1)
        def _repeat_solar_feature_over_pv_systems(solar_feature: torch.Tensor) -> torch.Tensor:
            # Check that the solar feature has been reshaped from `example time -> (example time)`.
            assert_num_dims(solar_feature, 1)
            return einops.repeat(
                solar_feature,
                "example -> example n_pv_systems 1",
                n_pv_systems=n_pv_systems,
            )

        solar_azimuth = _repeat_solar_feature_over_pv_systems(timeless_x[BatchKey.pv_solar_azimuth])
        solar_elevation = _repeat_solar_feature_over_pv_systems(
            timeless_x[BatchKey.pv_solar_elevation]
        )

        # Don't worry, we're not including "future" PV, because `reshape_time_as_batch`
        # zeros out the future timesteps.
        pv_power = timeless_x[BatchKey.pv].unsqueeze(-1)  # Shape: ((example * time) n_pv_systems 1)

        # The first element of dim 3 is zero for PV and one to mark that "this is GSP":
        pv_marker = torch.zeros_like(solar_azimuth)

        satellite_spacer = get_spacer_tensor(template=time_fourier, length=SATELLITE_SPACER_LEN)

        return torch.concat(
            (
                time_fourier,
                time_fourier_t0,  # So the model can tell which "step" this is.
                solar_azimuth,
                solar_elevation,
                y_fourier,
                x_fourier,
                satellite_spacer,
                pv_marker,
                pv_system_embedding,
                pv_power,
            ),
            dim=2,
        )


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
# for why we set `eq=False`
@dataclass(eq=False)
class GSPQueryGenerator(nn.Module):
    """Create a GSP query."""

    # This must be an InitVar because PyTorch does not allow modules to be
    # assigned before super().__init__()
    gsp_id_embedding: InitVar[nn.Embedding]

    def __post_init__(self, gsp_id_embedding):
        super().__init__()
        self.gsp_id_embedding = gsp_id_embedding

    def forward(
        self,
        x: dict[BatchKey, torch.Tensor],
        include_history: bool,
        base_batch_key: BatchKey,
        do_reshape_time_as_batch: bool,
    ) -> torch.Tensor:
        """Create query for GSP PV forecasts.

        Args:
            x: The batch. Requires BatchKeys: gsp, gsp_y_osgb_fourier, gsp_x_osgb_fourier,
                gsp_id, gsp_time_utc_fourier, solar_azimuth, gsp_t0_idx
            include_history: Whether or not to include GSP history up until t0_idx.
            base_batch_key: Either BatchKey.gsp or BatchKey.gsp_5_min.
            do_reshape_time_as_batch: If True, then reshape so each timestep is seen
                as a separate example.

        Returns tensor of shape (example * time, 1, query_dim) if do_reshape_time_as_batch,
            else (example, time, query_dim).
        """
        # The strategy is to get each tensor into the shape (example * time, 1, query_dim)
        # if `do_reshape_time_as_batch` else into the shape (example, time, query_dim).

        n_timesteps = x[base_batch_key].shape[1]

        def _get_batch_key(suffix: str) -> BatchKey:
            return BatchKey[base_batch_key.name + suffix]

        # Get tuple of BatchKeys to reshape each timestep as a new example:
        time_utc_fourier_batch_key = _get_batch_key("_time_utc_fourier")
        solar_az_batch_key = _get_batch_key("_solar_azimuth")
        solar_el_batch_key = _get_batch_key("_solar_elevation")

        t0_idx_batch_key = _get_batch_key("_t0_idx")
        t0_idx = x[t0_idx_batch_key]

        if do_reshape_time_as_batch:
            timeless_x = reshape_time_as_batch(
                x,
                batch_keys=(
                    base_batch_key,
                    time_utc_fourier_batch_key,
                    solar_az_batch_key,
                    solar_el_batch_key,
                ),
                set_to_zero_after_t0_idx=t0_idx if include_history else None,
                batch_keys_to_zero_out=[base_batch_key],
            )

        # gsp_{y,x}_osgb_fourier starts as shape (example, 1, fourier_features).
        y_fourier = x[BatchKey.gsp_y_osgb_fourier]
        x_fourier = x[BatchKey.gsp_x_osgb_fourier]

        # shape of time_fourier_t0: (example, 1, features)
        time_fourier_t0 = x[BatchKey.gsp_time_utc_fourier_t0].unsqueeze(1)

        gsp_id = x[BatchKey.gsp_id].squeeze()  # Shape: (example,)
        gsp_id_embedding = self.gsp_id_embedding(torch.nan_to_num(gsp_id, nan=0).int()).unsqueeze(1)
        # gsp_id_embedding is now shape (example, 1, dim)

        if do_reshape_time_as_batch:
            time_fourier = timeless_x[time_utc_fourier_batch_key].unsqueeze(1)
            solar_azimuth = timeless_x[solar_az_batch_key].unsqueeze(-1).unsqueeze(-1)
            solar_elevation = timeless_x[solar_el_batch_key].unsqueeze(-1).unsqueeze(-1)

            # Repeat across each timestep:
            y_fourier = y_fourier.repeat_interleave(repeats=n_timesteps, dim=0)
            x_fourier = x_fourier.repeat_interleave(repeats=n_timesteps, dim=0)
            gsp_id_embedding = gsp_id_embedding.repeat_interleave(repeats=n_timesteps, dim=0)
            time_fourier_t0 = time_fourier_t0.repeat_interleave(repeats=n_timesteps, dim=0)
        else:
            time_fourier = x[time_utc_fourier_batch_key]
            solar_azimuth = x[solar_az_batch_key].unsqueeze(-1)
            solar_elevation = x[solar_el_batch_key].unsqueeze(-1)

            # Repeat across each timestep:
            y_fourier = y_fourier.expand(-1, n_timesteps, -1)
            x_fourier = x_fourier.expand(-1, n_timesteps, -1)
            gsp_id_embedding = gsp_id_embedding.expand(-1, n_timesteps, -1)
            time_fourier_t0 = time_fourier_t0.expand(-1, n_timesteps, -1)

        # shape of time_fourier is now: (example * time) 1 features
        assert_num_dims(time_fourier, 3)

        # The first element of dim 3 is zero for PV and one to mark that "this is GSP":
        gsp_marker = torch.ones_like(solar_azimuth)

        satellite_spacer = get_spacer_tensor(template=time_fourier, length=SATELLITE_SPACER_LEN)

        gsp_query_tuple = (
            time_fourier,
            time_fourier_t0,
            solar_azimuth,
            solar_elevation,
            y_fourier,
            x_fourier,
            satellite_spacer,
            gsp_marker,
            gsp_id_embedding,
        )
        for value in gsp_query_tuple:
            print(value.shape)
        if include_history:
            if do_reshape_time_as_batch:
                # timeless_x already has the future zero'd out.
                print(timeless_x[base_batch_key].shape)
                history = timeless_x[base_batch_key].unsqueeze(-1).unsqueeze(-1)
            else:
                history = x[base_batch_key].detach().clone()
                history[:, t0_idx + 1 :] = 0
            print(history.shape)
            gsp_query_tuple += (history,)

        gsp_query = torch.concat(gsp_query_tuple, dim=2)
        print(gsp_query.shape)
        return gsp_query


def reshape_time_as_batch(
    x: dict[BatchKey, torch.Tensor],
    batch_keys: Sequence[BatchKey],
    set_to_zero_after_t0_idx: Optional[int] = None,
    batch_keys_to_zero_out: Optional[Sequence[BatchKey]] = None,
) -> dict[BatchKey, torch.Tensor]:
    new_batch: dict[BatchKey, torch.Tensor] = {}
    for batch_key in batch_keys:
        tensor = x[batch_key]
        if set_to_zero_after_t0_idx is not None and batch_key in batch_keys_to_zero_out:
            tensor = tensor.detach().clone()
            tensor[:, set_to_zero_after_t0_idx + 1 :] = 0
        new_batch[batch_key] = einops.rearrange(tensor, "example time ... -> (example time) ...")
    return new_batch



import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import (
    BatchKey,
    NumpyBatch,
    PreparedDataSource,
)
from power_perceiver.utils import datetime64_to_float

SAT_MEAN = {
    "HRV": 236.13257536395903,
    "IR_016": 291.61620182554185,
    "IR_039": 858.8040610176552,
    "IR_087": 738.3103442750336,
    "IR_097": 773.0910794778366,
    "IR_108": 607.5318145165666,
    "IR_120": 860.6716261423857,
    "IR_134": 925.0477987594331,
    "VIS006": 228.02134593063957,
    "VIS008": 257.56333202381205,
    "WV_062": 633.5975770915588,
    "WV_073": 543.4963868823854,
}

SAT_STD = {
    "HRV": 935.9717382401759,
    "IR_016": 172.01044433112992,
    "IR_039": 96.53756504807913,
    "IR_087": 96.21369354283686,
    "IR_097": 86.72892737648276,
    "IR_108": 156.20651744208888,
    "IR_120": 104.35287930753246,
    "IR_134": 104.36462050405994,
    "VIS006": 150.2399269307514,
    "VIS008": 152.16086321818398,
    "WV_062": 111.8514878214775,
    "WV_073": 106.8855172848904,
}


SATELLITE_CHANNEL_ORDER = ("example", "time", "channel", "y", "x")


def _set_sat_coords(dataset: xr.Dataset) -> xr.Dataset:
    """Set variables as coordinates"""
    return dataset.set_coords(
        ["time_utc", "channel_name", "y_osgb", "x_osgb", "y_geostationary", "x_geostationary"]
    )


class HRVSatellite(PreparedDataSource):
    sample_period_duration = pd.Timedelta("5 min")

    def process_before_transforms(self, dataset: xr.Dataset) -> xr.Dataset:
        # None of this will be necessary once this is implemented:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/629
        # Its backwards inteh prepared ones apparently
        dataset = dataset.reindex(y_geostationary_index=dataset.y_geostationary_index[::-1])
        # dataset = dataset.reindex(x=dataset.x[::-1])
        # Drop redundant coordinates (these are redundant because they
        # just repeat the contents of each *dimension*):
        dataset = dataset.drop_vars(
            [
                "example",
                "y_geostationary_index",
                "x_geostationary_index",
                "time_index",
                "channels_index",
            ]
        )

        # Rename coords to be more explicit about exactly what some coordinates hold:
        dataset = dataset.rename_vars(
            {
                "channels": "channel_name",
                "time": "time_utc",
            }
        )

        # Rename dimensions. Standardize on the singular (time, channel, etc.).
        # Remove redundant "index" from the dim name. These are *dimensions* so,
        # by definition, they are indicies!
        dataset = dataset.rename_dims(
            {
                "y_geostationary_index": "y",
                "x_geostationary_index": "x",
                "time_index": "time",
                "channels_index": "channel",
            }
        )
        # Setting coords won't be necessary once this is fixed:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/627
        dataset = _set_sat_coords(dataset)

        dataset = dataset.transpose(*SATELLITE_CHANNEL_ORDER)
        dataset.attrs["t0_idx"] = self.t0_idx
        return dataset

    @staticmethod
    def to_numpy(dataset: xr.Dataset) -> NumpyBatch:
        """This is called from Dataset.__getitem__.

        This processes this modality's xr.Dataset, to convert the xr.Dataset
        into a dictionary mapping BatchKeys to numpy arrays, as documented
        in the BatchKey class.
        """
        batch: NumpyBatch = {}

        # Prepare the satellite imagery itself
        hrvsatellite = dataset["data"]
        # hrvsatellite is int16 on disk
        hrvsatellite = hrvsatellite.astype(np.float32)
        hrvsatellite -= SAT_MEAN["HRV"]
        hrvsatellite /= SAT_STD["HRV"]
        batch[BatchKey.hrvsatellite_actual] = hrvsatellite.values
        # TODO: Double check it should be float32 or int8?
        #  It seems undone in rawSatellite, so should work

        # Coordinates
        batch[BatchKey.hrvsatellite_t0_idx] = dataset.attrs["t0_idx"]
        batch[BatchKey.hrvsatellite_time_utc] = datetime64_to_float(dataset["time_utc"].values)
        for batch_key, dataset_key in (
            (BatchKey.hrvsatellite_y_osgb, "y_osgb"),
            (BatchKey.hrvsatellite_x_osgb, "x_osgb"),
            (BatchKey.hrvsatellite_y_geostationary, "y_geostationary"),
            (BatchKey.hrvsatellite_x_geostationary, "x_geostationary"),
        ):
            # HRVSatellite coords are already float32.
            batch[batch_key] = dataset[dataset_key].values

        return batch



import datetime

import pandas as pd
import pytest

from power_perceiver.consts import Location
from power_perceiver.load_raw.data_sources.raw_satellite_data_source import RawSatelliteDataSource

# for the moment
SAT_HEIGHT_IN_PIXELS = 128
SAT_WIDTH_IN_PIXELS = 256


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_init(sat_data_source):
    assert sat_data_source._data_in_ram is None
    assert sat_data_source._data_on_disk is None


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_init_start_and_end_dates_swapped(get_sat_data_source):
    with pytest.raises(AssertionError):
        get_sat_data_source(
            end_date=datetime.datetime(year=2020, month=1, day=1),
            start_date=datetime.datetime(year=2020, month=12, day=31, hour=23, minute=59),
        )


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_per_worker_init(sat_data_opened):
    assert sat_data_opened._data_on_disk is not None


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_datetimes(sat_data_opened):
    dt_index = sat_data_opened.datetime_index
    # Test that we don't have any datetimes at midnight (00:00 to 00:59):
    assert not (dt_index.hour == 0).any()


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_get_spatial_slice(sat_data_opened: RawSatelliteDataSource):
    # Select a location roughly in the middle of the Satellite imagery:
    location_center_osgb = Location(x=66400, y=357563)
    selection = sat_data_opened._get_spatial_slice(
        sat_data_opened._data_on_disk,
        location_center_osgb,
    )
    assert len(selection.x_geostationary) == SAT_WIDTH_IN_PIXELS
    assert len(selection.y_geostationary) == SAT_HEIGHT_IN_PIXELS

    with pytest.raises(AssertionError):
        selection = sat_data_opened._get_spatial_slice(
            sat_data_opened._data_on_disk,
            # Try getting an image centered on the left edge:
            Location(x=-1371057, y=357563),
        )


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_load_subset_into_ram(sat_data_loaded: RawSatelliteDataSource):
    assert sat_data_loaded._data_in_ram is not None


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_get_example(sat_data_loaded: RawSatelliteDataSource):
    xr_example = sat_data_loaded.get_example(
        t0_datetime_utc=pd.Timestamp("2020-01-01 12:00"),
        center_osgb=Location(x=66400, y=357563),
    )
    del xr_example  # TODO: Do something with this!


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_get_osgb_location_for_example(sat_data_loaded: RawSatelliteDataSource):
    location = sat_data_loaded.get_osgb_location_for_example()
    print("LOCATION!")
    print(location)
    for coord in location:
        assert isinstance(coord, float), f"{type(coord)=}"
    # TODO: Check the OSGB coords are sane!



import datetime

import matplotlib
import matplotlib.dates as mdates
import torch
import torch.nn.functional as F
from torch import nn


class MixtureDensityNetwork(nn.Module):
    def __init__(self, in_features: int, num_gaussians: int = 2):
        super().__init__()
        self.pi = nn.Linear(in_features=in_features, out_features=num_gaussians)
        self.mu = nn.Linear(in_features=in_features, out_features=num_gaussians)
        self.sigma = nn.Linear(in_features=in_features, out_features=num_gaussians)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pi = self.pi(x)
        pi = F.softmax(pi, dim=-1)
        mu = self.mu(x)
        # mu = F.relu(mu)  # Guarantee that PV can't go negative!
        sigma = self.sigma(x)
        sigma = torch.exp(sigma)
        return torch.concat((pi, mu, sigma), dim=-1)


def get_pi_mu_sigma(
    network_output: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    assert network_output.shape[-1] % 3 == 0
    num_gaussians = int(network_output.shape[-1] / 3)
    pi = network_output[..., :num_gaussians]
    mu = network_output[..., num_gaussians : num_gaussians * 2]
    sigma = network_output[..., num_gaussians * 2 :]
    return pi, mu, sigma


def get_distribution(network_output: torch.Tensor) -> torch.distributions.MixtureSameFamily:
    pi, mu, sigma = get_pi_mu_sigma(network_output)

    mixture_distribution = torch.distributions.Categorical(probs=pi)
    component_distribution = torch.distributions.Normal(loc=mu, scale=sigma)
    gaussian_mixture_model = torch.distributions.MixtureSameFamily(
        mixture_distribution, component_distribution
    )
    return gaussian_mixture_model


def plot_probs(
    ax: matplotlib.axes.Axes,
    network_output: torch.Tensor,
    left: datetime.datetime,
    right: datetime.datetime,
) -> matplotlib.axes.Axes:
    """Plot distribution over time.

    Args:
        network_output: Shape: (time, mixture_density_network_features)
        left, right: Pass in the output of `mdates.date2num` for the left and
            right time boundaries.
    """
    left = mdates.date2num(left)
    right = mdates.date2num(right)

    sweep_n_steps = 100
    sweep_start = 1
    sweep_stop = 0

    n_timesteps = network_output.shape[0]

    # Define a 'sweep' matrix which we pass into log_prob to get probabilities
    # for a range of values at each timestep. Those values range from sweep_start to sweep_stop.
    sweep = torch.linspace(
        start=sweep_start,
        end=sweep_stop,
        steps=sweep_n_steps,
        dtype=torch.float32,
        device=network_output.device,
    )
    sweep = sweep.unsqueeze(-1).expand(sweep_n_steps, n_timesteps)

    # Get probabilities.
    distribution = get_distribution(network_output)
    log_probs = distribution.log_prob(sweep)
    probs = torch.exp(log_probs).detach().cpu().numpy()

    # Normalise so every column has a max of 1 otherwise blurry distributions are very hard to see!
    probs /= probs.max(axis=0)

    # Plot!
    extent = (left, right, sweep_stop, sweep_start)  # left, right, bottom, top
    ax.imshow(probs, aspect="auto", interpolation="none", extent=extent, cmap="Greys")

    return ax
